//
// Copyright (c) 2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_EXECUTOR_H_INCLUDED__
#define __WZCPP_EXECUTOR_H_INCLUDED__

#include "wzcpp/features/logging.h"
#include <atomic>
#include <condition_variable>
#include <functional>
#include <future>
#include <mutex>
#include <queue>
#include <thread>
#include <utility>

namespace wzcpp
{

/*!
@brief Baseclass for an queued executor

 */
class ExecutorBase
{
public:
	using TaskFunction = std::function<void()>;
	using TaskQueue    = std::queue<TaskFunction>;
	using UniqueLock   = std::unique_lock<std::mutex>;

public:
	ExecutorBase()               = default;
	// Moving is prevented because the conditions_variable is not moveable
	ExecutorBase(ExecutorBase&&) = delete;
	ExecutorBase& operator=(ExecutorBase&& source) = delete;
	// *********************** Make non copyable ***************************
	ExecutorBase(const ExecutorBase& source) = delete;
	ExecutorBase& operator=(const ExecutorBase& source) = delete;

	virtual ~ExecutorBase()
	{
		wzlogs_d("Destroying BaseExecutor object");
		stop();
	}

public:
	/*!
	@brief Accepts any callable function with no parameters and no return value, the function must be noexcept

	@tparam SimpleFunc
	@param function
	@pre the function must be noexcept
	 */
	template<typename SimpleFunc>
	void postAndForget(SimpleFunc&& function)
	{
		wzlogs_d("Scheduled new function. Currently free slots: " << getFreeSlots());
		// Add to task queue
		{
			std::lock_guard<std::mutex> lk(dataMutex_);
			if (shouldStop_ || shuttingDown_)
				throw std::runtime_error(
				    "BaseExecutor::postAndForget() - attempted to add task to stopped BaseExecutor");
			// Add our task to the queue
			taskQueue_.emplace(std::forward<SimpleFunc>(function));
		}

		// Notify about new work
		cv_.notify_one();
	}

	/*!
	@brief Accepts any callable with any return type and variable amount and types of parameters

	@tparam Function any callable type
	@tparam Args the arguments for the callable
	@param function the acutal function pointer for the task to run
	@param args the actual parameters for the task callable
	@return a future of the type that template parameter Function returns
	 */
	template<typename Function, typename... Args>
	auto post(Function&& function, Args&&... args)

	    -> std::future<typename std::invoke_result_t<Function, Args...>>
	{
		// Types
		using return_t = typename std::invoke_result_t<Function, Args...>;

		// Create packaged task
		auto task = std::make_shared<std::packaged_task<return_t()>>(
		    std::bind(std::forward<Function>(function), std::forward<Args>(args)...));

		std::future<return_t> result = task->get_future();

		// Add task to queue
		{
			UniqueLock lk(dataMutex_);

			if (shouldStop_ || shuttingDown_)
				throw std::runtime_error(
				    "ERROR: Thread::ThreadPool::Add_Task() - attempted to add task to stopped BaseExecutor");

			// Add our task to the queue
			taskQueue_.emplace([task]() { (*task)(); });
		}

		// Notify that there is work to do
		cv_.notify_one();

		return result;
	}

	/*!
	@brief Shutdown finshes all queued task, no new tasks will be accepted

	 */
	void shutdown()
	{
		wzlogs_d("Shutting down the executor");
		pause(false);  // unpause if paused
		shuttingDown_ = true;
		cv_.notify_all();
		wzlogs_d("Shutting down the executor done");
	}

	/*!
	@brief Stopping only finishes the currently running tasks, no new queued tasks will be executed. No new tasks will
	be accepted in this state.

	 */
	void stop()
	{
		wzlogs_d("Stopping the executor");
		shouldStop_ = true;
		pause(false);  // unpause if paused

		cv_.notify_all();
		wzlogs_d("Stopping the executor done");
	}

	void pause(bool pauseState)
	{
		bool hasChanged{false};
		wzlogs_d("Setting pause to " << std::boolalpha << pauseState);
		{
			UniqueLock lk(dataMutex_);
			if (isPaused_ != pauseState)
			{
				isPaused_  = pauseState;
				hasChanged = true;
			}
		}
		if (hasChanged)
		{
			cv_.notify_all();
		}
	}

	virtual void init()
	{
		TaskQueue empty;
		std::swap(taskQueue_, empty);
		shouldStop_   = false;
		shuttingDown_ = false;
	}


	virtual auto getFreeSlots() -> std::size_t = 0;
	virtual auto areTasksRunning() -> bool     = 0;

	auto waitForFinished() -> void
	{

		wzlogs_d("Waiting for all tasks to complete");
		UniqueLock lk{finishedMutex_};
		UniqueLock dlk{dataMutex_};
		if (isPaused_)
		{
			wzlogs_w("Warning waiting on a paused executor");
		}
		dlk.unlock();
		cvFinished_.wait(lk, [this]() { return ((taskQueue_.empty() && !areTasksRunning()) || shouldStop_); });
		if (shouldStop_)
		{
			wzlogs_d("waitForFinished executor stopped");
		}
		else
		{
			wzlogs_d("Taskqueue empty and no tasks are running");
		}
	}

protected:
	TaskQueue taskQueue_;
	std::condition_variable cv_;
	std::condition_variable cvFinished_;
	std::atomic<bool> shouldStop_{true};
	std::atomic<bool> shuttingDown_{true};

	bool isPaused_{false};
	mutable std::mutex dataMutex_;
	mutable std::mutex finishedMutex_;

private:
	// Logging
	static constexpr const char* const logName_ = "ExecutorBase";
	WZLOGGER(logName_, LVL_DEBUG);
};


/*!
@brief a derived class working with a pool of std::threads

 */
class ThreadPoolExecutor : public ExecutorBase
{
public:
	using Threads = std::vector<std::thread>;

	void init() override
	{
		if (!shouldStop_ && !shuttingDown_)
		{
			wzlogs_d("Trying to init a not stopped executor");
			stop();
		}
		wzlogs_d("threads size: " << threads_.size() << " used Threads:" << usedThreads_);
		for (auto& taskThread : threads_)
		{
			if (taskThread.joinable())
			{
				wzlogs_d("Waiting for thread " << taskThread.get_id() << " to join " << taskThread.joinable());
				taskThread.join();
				wzlogs_d("Thread " << taskThread.get_id() << " joined");
			}
		}
		wzlogs_d("All threads joined");
		ExecutorBase::init();

		threads_.clear();
		for (std::size_t i = 0; i < threadCount_; ++i)
		{
			threads_.emplace_back([this]() { taskRunner(); });
		}
	}

private:
	Threads threads_;
	std::size_t threadCount_;

public:
	ThreadPoolExecutor()

	    : ThreadPoolExecutor(std::thread::hardware_concurrency())
	{
		init();
	}

	ThreadPoolExecutor(const std::size_t threadCount)
	{
		// Sanity
		if (threadCount == 0) throw std::logic_error("ThreadPool() -- must have at least one thread");
		threadCount_ = threadCount;
		threads_.reserve(threadCount);
		init();
	}

	// *********************** Make non copyable ***************************
	ThreadPoolExecutor(const ThreadPoolExecutor& source) = delete;
	ThreadPoolExecutor& operator=(const ThreadPoolExecutor& source) = delete;

	virtual ~ThreadPoolExecutor()
	{
		wzlogs_d("Destroying ThreadPoolExecutor object ");
		stop();
		wzlogs_d("Destroying ThreadPoolExecutor object: lock aquired ");
		for (auto& taskThread : threads_)
		{
			if (taskThread.joinable())
			{
				taskThread.join();
			}
		}
		wzlogs_d("All threads joined");
	}

	decltype(threads_.size()) getThreadCount() const
	{
		return threads_.size();
	}

	auto getUsedThreads() -> std::size_t
	{
		std::lock_guard<std::mutex> lk(dataMutex_);
		return usedThreads_;
	}

	auto getFreeSlots() -> std::size_t override
	{
		UniqueLock lk(dataMutex_);
		// Available threads - currently used threads - amount of tasks in queue = how
		// many tasks can be run directly without letting them wait for a free thread
		auto result = getThreadCount() - usedThreads_;  // free threads
		if (result == 0 || result <= taskQueue_.size())
		{
			result = 0;
		}
		else
		{
			result -= taskQueue_.size();
		}

		return result;
	}

	auto areTasksRunning() -> bool override
	{
		return (getThreadCount() > getFreeSlots());
	}

private:
	void taskRunner()
	{
		const std::thread::id this_id = std::this_thread::get_id();
		wzlogs_d("taskRunner started " << this_id);
		while (!shouldStop_)
		{

			// Scoped waiting / task-retrieval block
			{
				// Wait on tasks or 'stop processing' flags
				UniqueLock lk(dataMutex_);

				wzlogs_d("Waiting for work " << this_id);
				auto predicate = [this]() {
					return ((!taskQueue_.empty() && !isPaused_) || shouldStop_ || shuttingDown_);
				};
				if (!predicate())
				{
					cv_.wait(lk, predicate);
				}
				wzlogs_d("Done with waiting " << this_id);

				if (shouldStop_ || (taskQueue_.empty() && shuttingDown_))
				{
					break;
				}
				// Retrieve next task
				auto task = std::move(taskQueue_.front());
				taskQueue_.pop();

				wzlogs_d("taskRunner executing task " << this_id);
				usedThreads_++;
				wzlogs_d("taskRunner used threads " << usedThreads_);
				lk.unlock();
				task();
				lk.lock();
				if (usedThreads_ > 0)
				{
					usedThreads_--;
				}
				if (taskQueue_.empty())  // The queue is empty and we are done with the current task
				{
					if (usedThreads_ == 0)  // no other thread is running a task-> this thread is the last one running
					{
						cvFinished_.notify_all();
					}
					if (shuttingDown_)
					{
						break;
					}
				}
				wzlogs_d("taskRunner used threads " << usedThreads_);
			}
		}

		wzlogs_d("taskRunner finished " << this_id);
	}

private:
	// Logging
	static constexpr const char* const logName_ = "ThreadPoolExecutor";
	WZLOGGER(logName_, LVL_DEBUG);
	std::uint32_t usedThreads_{0};
};
}  // namespace wzcpp

#endif  // __WZCPP_EXECUTOR_H_INCLUDED__