//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_ERROR_EXCEPTION_H__
#define __WZCPP_ERROR_EXCEPTION_H__

#include <cstring>
#include <sstream>
#include <stdexcept>


// clang-format off
#define DECLARE_WZ_EXCEPTION(CLSNAME, ERRORCODE, BASE)														\
	class  CLSNAME : public BASE																			\
	{																										\
	public:																									\
		CLSNAME(wzcpp::error::SourceLineInfo&& sli, const char* userMsg) : BASE(#CLSNAME, static_cast<std::uint32_t>(ERRORCODE), std::move(sli), userMsg)	\
		{																										\
		}																										\
		CLSNAME(wzcpp::error::SourceLineInfo&& sli, std::string&& userMsg)						 							\
			: BASE(#CLSNAME, static_cast<std::uint32_t>(ERRORCODE),std::move(sli), std::move(userMsg))			\
		{																										\
		}																										\
		CLSNAME(wzcpp::error::SourceLineInfo&& sli, std::string&& userMsg, std::uint32_t userErrorCode)						\
			: BASE(#CLSNAME, static_cast<std::uint32_t>(ERRORCODE), std::move(sli),std::move(userMsg), userErrorCode) \
		{																										\
		}																										\
		CLSNAME(wzcpp::error::SourceLineInfo&& sli, const char* userMsg, std::uint32_t userErrorCode)							\
			: BASE(#CLSNAME, static_cast<std::uint32_t>(ERRORCODE), std::move(sli),userMsg, userErrorCode)		\
		{																										\
		}																										\
		CLSNAME(const char* msg, std::uint32_t wzErrorCode, wzcpp::error::SourceLineInfo&& sli, const char* userMsg) : BASE(msg, wzErrorCode, std::move(sli), userMsg) \
		{																										\
		}																										\
		CLSNAME(const char* msg, std::uint32_t wzErrorCode, wzcpp::error::SourceLineInfo&& sli, const char* userMsg, std::uint32_t userErrorCode)	\
			: BASE(msg, wzErrorCode, std::move(sli), userMsg, userErrorCode)														\
		{																										\
		}																										\
		CLSNAME(const char* msg, std::uint32_t wzErrorCode, wzcpp::error::SourceLineInfo&& sli, std::string&& userMsg)		\
			: BASE(msg, wzErrorCode, std::move(sli), std::move(userMsg))										\
		{																										\
		}																										\
		CLSNAME(const char* msg, std::uint32_t wzErrorCode, wzcpp::error::SourceLineInfo&& sli, std::string&& userMsg, std::uint32_t userErrorCode)	\
			: BASE(msg, wzErrorCode, std::move(sli), std::move(userMsg), userErrorCode)							\
		{																										\
		}																										\
	}
 
// Short macro to create SourcelineInfo from current line
#define WZ_SL  wzcpp::error::SourceLineInfo(__FILE__, __LINE__)
// clang-format on

namespace wzcpp
{
namespace error
{
struct SourceLineInfo
{
	const char* file_;
	std::size_t line_;

	SourceLineInfo() : file_(""), line_(0){};
	SourceLineInfo(const char* file, std::size_t line) : file_(file), line_(line){};

	bool empty() const
	{
		return file_[0] == '\0';
	}

	bool operator==(SourceLineInfo const& other) const
	{
		return line_ == other.line_ && (file_ == other.file_ || std::strcmp(file_, other.file_) == 0);
	}

	bool operator<(SourceLineInfo const& other) const
	{
		return line_ < other.line_ || (line_ == other.line_ && (std::strcmp(file_, other.file_) < 0));
	}
};

class WzBaseException : public std::exception
{
public:
	WzBaseException(const char* msg, std::uint32_t wzErrorCode, SourceLineInfo&& lineInfo)
	    : userMsg_(""),
	      userErrorCode_{0},
	      msg_{msg},
	      wzErrorCode_{wzErrorCode},
	      sourceLineInfo{lineInfo}
	{
	}

	WzBaseException(const char* msg, std::uint32_t wzErrorCode, SourceLineInfo&& lineInfo, const char* userMsg)
	    : userMsg_(userMsg),
	      userErrorCode_{0},
	      msg_{msg},
	      wzErrorCode_{wzErrorCode},
	      sourceLineInfo{lineInfo}
	{
	}

	WzBaseException(const char* msg,
	                std::uint32_t wzErrorCode,
	                SourceLineInfo&& lineInfo,
	                const char* userMsg,
	                std::uint32_t userErrorCode)
	    : userMsg_(userMsg),
	      userErrorCode_{userErrorCode},
	      msg_{msg},
	      wzErrorCode_{wzErrorCode},
	      sourceLineInfo{lineInfo}
	{
	}

	WzBaseException(const char* msg, std::uint32_t wzErrorCode, SourceLineInfo&& lineInfo, std::string&& userMsg)
	    : userMsg_(std::move(userMsg)),
	      userErrorCode_{0},
	      msg_{msg},
	      wzErrorCode_{wzErrorCode},
	      sourceLineInfo{lineInfo}
	{
	}

	WzBaseException(const char* msg,
	                std::uint32_t wzErrorCode,
	                SourceLineInfo&& lineInfo,
	                std::string&& userMsg,
	                std::uint32_t userErrorCode)
	    : userMsg_(std::forward<std::string>(userMsg)),
	      userErrorCode_{userErrorCode},
	      msg_{msg},
	      wzErrorCode_{wzErrorCode},
	      sourceLineInfo{lineInfo}
	{
	}

	virtual ~WzBaseException() = default;

	const char* what() const throw()
	{
		return msg_;
	}

	virtual std::string msg() const throw()
	{
		std::ostringstream oss;
		oss << "[F:" << sourceLineInfo.file_ << ";L:" << sourceLineInfo.line_ << "](" << wzErrorCode_ << ")"
		    << ": " << userMsg_ << " UserErrorCode:" << userErrorCode_;
		return oss.str().c_str();
	}

	const SourceLineInfo& getSourceLine() const
	{
		return sourceLineInfo;
	}

	std::string getUserMsg() const
	{
		return userMsg_;
	}

	std::uint32_t getWzErrorCode() const
	{
		return wzErrorCode_;
	}

	std::uint32_t getErrorCode() const
	{
		return userErrorCode_;
	}

protected:
	std::string userMsg_;
	const std::uint32_t userErrorCode_;
	const char* msg_{"WzBaseException"};
	const std::uint32_t wzErrorCode_{0};
	const SourceLineInfo sourceLineInfo;
};

enum class ErrorCodes
{
	NotImplemented = 1,
	OutOfRange     = 2
};

DECLARE_WZ_EXCEPTION(
    NotImplementedException,
    ErrorCodes::NotImplemented,
    WzBaseException);  // Exception class that can be used if a function is not implemented and should throw an error


}  // namespace error
}  // namespace wzcpp
#endif  // !__WZCPP_ERROR_EXCEPTION_H__

