//
// Copyright (c) 2002-2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_TIMER_H__
#define __WZCPP_TIMER_H__

#include "wzcpp/features/logging.h"

#include <atomic>
#include <condition_variable>
#include <cstdint>
#include <ctime>
#include <deque>
#include <iomanip>
#include <map>
#include <mutex>
#include <numeric>
#include <sstream>

namespace wzcpp
{

// Alias
using Duration  = std::chrono::duration<std::int64_t>;
using TimePoint = std::chrono::system_clock::time_point;

constexpr auto IDLE_TIMER_INTERVAL = std::chrono::seconds(60);

// helper function

/*!
Convert a system_clock time point to a string
*/
inline std::string toString(const std::chrono::system_clock::time_point& tp)
{
	std::ostringstream oss;
	std::time_t t = std::chrono::system_clock::to_time_t(tp);
	std::tm tm    = *std::localtime(&t);
	oss << std::put_time(&tm, "%c %Z");
	return oss.str();
}


enum class TimerType
{
	Unknown,
	Periodic,
	OneShot
};

using TimerCallback = std::function<void()>;

/*!
@brief A structure holding information about a timer
 */
struct TimerInfo
{
	std::uint64_t id{0};    /*!< The unique id of the timer */
	Duration duration;      /*!< The duration of one timer intervall */
	std::string name;       /*!< The name of the timer, helps to identify it */
	TimerType type;         /*!< The type of the timer, oneshot or periodic */
	TimerCallback callback; /*!< A callable to invoke if the timer is expired */
};

template<class clock>
class TimerManager;

/*!
@brief Represents a timer object and provides functions to interact with it

@tparam clock the underlying clock type like steady_clock or system_clock on whitch the timer operates
 */
template<class clock>
class Timer
{
public:
	Timer(Duration interval, TimerManager<clock>& tm, const std::string& name = "") : manager_{tm}

	{
		info_.id       = tm.getNextFreeTimerId();
		info_.duration = interval;
		info_.name     = name;
		info_.callback = [this]() { this->onTimer(); };
	};

	~Timer()
	{
		std::lock_guard<std::mutex> lk(mutex_);
		if (info_.id > 0)
		{
			manager_.remove(info_.id);
		}
	}

	/*!
	@brief The timer is scheduled to run once and after it it expired the timers onTimer callback will be invoked and
	the generic callback of the timermanger for all timers is also called.
	@throws Throws an logic_error if the timer is already running
	 */
	auto runOnce() -> void
	{
		std::lock_guard<std::mutex> lk(mutex_);
		if (!running_ && info_.id > 0)
		{
			info_.type = TimerType::OneShot;
			manager_.schedule(info_);
			running_ = true;
		}
		else
		{
			std::ostringstream oss;
			oss << "Timer already running or invalid id: " << info_.id;
			wzlogs_e(oss.str());
			throw std::logic_error(oss.str());
		}
	}

	/*!
	@brief The timer is scheduled to periodic and after every run the timers onTimer callback will be invoked and the
	generic callback of the timermanger for all timers is also called. To stop the timer the cancel function may be
	used.
	@see cancel
	@throws Throws an logic_error if the timer is already running
	 */
	auto runPeriodic() -> void
	{
		std::lock_guard<std::mutex> lk(mutex_);
		if (!running_ && info_.id > 0)
		{
			wzlogs_d("Starting timer " << info_.id << "(" << info_.name << ")"
			                           << " periodic");
			info_.type = TimerType::Periodic;
			manager_.schedule(info_);
			running_ = true;
		}
		else
		{
			std::ostringstream oss;
			oss << "Timer already running or invalid id: " << info_.id;
			wzlogs_e(oss.str());
			throw std::logic_error(oss.str());
		}
	}

	/*!
	@brief Get the timer id

	@return std::uint64_t the unique timer id or 0 if the timer is not valid
	 */
	auto getId() -> std::uint64_t
	{
		std::lock_guard<std::mutex> lk(mutex_);
		return info_.id;
	}

	/*!
	@brief Get the timespan of the timer

	@return Duration
	 */
	auto getInterval() -> Duration
	{
		std::lock_guard<std::mutex> lk(mutex_);
		return info_.duration;
	}

	/*!
	@brief Get the timer name

	@return std::string the name of the timer as string
	*/
	auto getName() -> std::string
	{
		std::lock_guard<std::mutex> lk(mutex_);
		return info_.name;
	}

	/*!
	@brief Cancels a running timer

	@return true if the timer was running and is now canceled, false if it was not running at all
	 */
	auto cancel() -> bool
	{

		bool result{false};
		if (running_)
		{
			std::lock_guard<std::mutex> lk(mutex_);
			result   = manager_.cancel(info_.id);
			running_ = false;
			wzlogs_d("Canceled timer with id " << info_.id << ", manager returned: " << std::boolalpha << result);
		}
		return result;
	}

	/*!
	@brief Completely removes a timer from the timermanager. After that the timer is invalid and cannot be used anymore

	@return true if the timer was running and is now removed, false if it was not running at all
	 */
	auto remove() -> bool
	{
		std::lock_guard<std::mutex> lk(mutex_);
		const bool result = manager_.remove(info_.id);
		running_          = false;
		info_.id          = 0;
		return result;
	}

	/*!
	@brief Resets a timer so that new starting point is now.

	 */
	auto reset() -> void
	{
		std::lock_guard<std::mutex> lk(mutex_);
		manager_.cancel(info_.id);
		manager_.schedule(info_);
	}

	auto isRunning() -> bool
	{
		return running_;
	}

	/*!
	@brief callback function, which is invoked of the timer has expired
	@remark The function runs in the context of the timer thread.

	 */
	auto onTimer() noexcept -> void
	{
		wzlogs_d("Timer" << info_.id << " on Timer callback invoked. Running:" << running_);
		{
			std::lock_guard<std::mutex> lk(mutex_);
			if (running_ && info_.type == TimerType::Periodic)
			{
				manager_.schedule(info_);
			}
			else
			{
				running_ = false;
			}
		}
		std::lock_guard<std::mutex> lk(callbackMutex_);
		if (userCallback_)
		{
			userCallback_();
		}
	}

	auto setCallback(TimerCallback callback) -> void
	{
		std::lock_guard<std::mutex> lk(callbackMutex_);
		userCallback_ = callback;
	}

protected:
	// Logging
	static constexpr const char* const logName_ = "Timer";
	WZLOGGER(logName_, LVL_DEBUG);

	mutable std::mutex mutex_;
	mutable std::mutex callbackMutex_;

	std::atomic<bool> running_{false};
	TimerManager<clock>& manager_;
	TimerInfo info_;
	TimerCallback userCallback_;
};

/*!
@brief The timer manager is responsible to handle the single timers.

@tparam clock the underlying clock type like steady_clock or system_clock on whitch the timermanager operates. It has to
be the same for the timer which are managed by this class.
 */
template<class clock>
class TimerManager
{
public:
	using TimerIdList       = std::vector<std::uint64_t>;
	using AllTimersCallback = std::function<void(TimerIdList)>;
	using UniqueLock        = std::unique_lock<std::mutex>;

	TimerManager();
	TimerManager(TimerManager&&) = default;
	~TimerManager()              = default;

	TimerManager& operator=(TimerManager&&) = default;

	// Make non copyable
	TimerManager(const TimerManager&) = delete;
	TimerManager& operator=(const TimerManager&) = delete;


	/*!
	@brief Schedules a timer for later execution

	@param ti The information struct for the timer to schedule
	 */
	auto schedule(const TimerInfo& ti) -> void;

	/*!
	@brief Cancles a timer iditified by its unique id

	@param id the unique timer id
	@return true if the timer was running and is now canceled, false if it was not running at all
	 */
	auto cancel(std::uint64_t id) -> bool;

	/*!
	@brief Removes a timer identified by its unique id

	@param id  the unique timer id
	@return true if the timer was running and is now removed, false if it was not running at all
	 */
	auto remove(std::uint64_t id) -> bool;
	auto getNextFreeTimerId() -> std::uint64_t; /*!< Returns the next unique id, for a timer */

	/*!
	@brief  Set the callback which should be invoked for any expired timer

	@param callback The callback function which must be noexcept
	 */
	auto setTimersCallback(AllTimersCallback callback) -> void;

	auto init() -> void;  // Initializes the TimerManager, normally done in constructor but if a  TimerManager object is
	                      // reused this function has to be called manually
	auto run() -> void;   /*!< The main function for the timer service, this will block until the service is stopped. */
	auto shutdown() -> void;        /*!< Stops the timer service and cancels the run function */
	auto waitForFinished() -> void; /*!< Blocks till the run function has completed */

private:
	/*!
	@brief Notifies the timer services, that something with the timers has changed

	    */
	auto notifyReschedule() -> void
	{
		wzlogs_d("Notify reschedule");
		{
			std::lock_guard<std::mutex> lk(cv_mut_);
			reschedule_ = true;
		}
		cv_.notify_one();
	}

	auto idValidGuard(std::uint64_t id) -> void
	{
		// Check if id != 0 and smaller than the maximum poolSize_, also make sure that the id is not part
		// of the available ids in the pool
		if (std::find(idPool_.begin(), idPool_.end(), id) != idPool_.end() || id == 0 || id > poolSize_)
		{
			std::ostringstream oss;
			oss << "Invalid Id " << id << " in schedule of TimeManager";
			wzlogs_e(oss.str());
			throw std::logic_error(oss.str());
		}
	}

private:
	// Logging
	static constexpr const char* const logName_ = "TimerManager";
	WZLOGGER(logName_, LVL_DEBUG);

	// Id pool for timer ids
	static constexpr std::uint32_t poolSize_{1000};
	std::deque<std::uint64_t> idPool_;

	mutable std::mutex timer_mut_;
	std::map<typename clock::time_point, TimerInfo> timers_;

	std::condition_variable cv_;
	mutable std::mutex cv_mut_;

	std::condition_variable cvFinished_;
	mutable std::mutex cvFinished_mut_;
	std::atomic<bool> stopped_;

	bool reschedule_{false};

	std::atomic<bool> shouldStop_{false};

	mutable std::mutex cb_mut_;
	AllTimersCallback multiCallback_;  // callback for multiple timers
};


/*********************************************************
 * Implementation TimerManager
 *********************************************************/
template<class clock>
TimerManager<clock>::TimerManager() : idPool_(poolSize_),
                                      stopped_{true}
{
	std::iota(idPool_.begin(), idPool_.end(), 1);  // Fill the id pool with ids
	std::lock_guard<std::mutex> lk(cv_mut_);
	reschedule_ = false;
}

template<class clock>
auto TimerManager<clock>::schedule(const TimerInfo& ti) -> void
{
	const auto tp = clock::now() + ti.duration;  // Calculate the next timestamp for the timer to schedule

	idValidGuard(ti.id);

	wzlogs_d("Schedule new wake up at " << toString(tp) << " of timer " << ti.name << "(" << ti.id
	                                    << ")");  // DEBUG information
	auto ms = std::chrono::time_point_cast<std::chrono::milliseconds>(tp).time_since_epoch().count();
	wzlogs_d("In ms: " << ms);

	{  // scope for the mutex lock of timer_mut_
		std::lock_guard<std::mutex> lk(timer_mut_);
		auto result = timers_.emplace(tp, ti);  // Add timer to internal timers_ map
		if (result.second != true)  // if there was already an entry with that exact timepoint throw exception
		{
			const auto msg = "There is already a timer scheduled at this timepoint";
			wzlogs_e(msg);
			throw std::logic_error(msg);
		}
	}
	notifyReschedule();  // Notify the timermanager that the scheduled timers have changed
}

template<class clock>
auto TimerManager<clock>::cancel(std::uint64_t id) -> bool
{
	wzlogs_d("Cancel timer with id " << id);
	std::unique_lock<std::mutex> lk(timer_mut_);
	idValidGuard(id);
	bool result{false};

	for (auto it = timers_.begin(); it != timers_.end();
	     /* no increment */)
	{
		if (it->second.id == id)
		{
			result = true;
			wzlogs_d("Canceled timer");
			it = timers_.erase(it);
		}
		else
		{
			++it;
		}
	}
	if (result)
	{
		lk.unlock();
		notifyReschedule();  // Notify the timermanager that the scheduled timers have changed
	}
	return result;
}

template<class clock>
auto TimerManager<clock>::remove(std::uint64_t id) -> bool
{
	wzlogs_d("Remove timer with id " << id);
	bool result = cancel(id);
	std::lock_guard<std::mutex> lk(timer_mut_);
	idPool_.push_front(id);
	return result;
}

template<class clock>
auto TimerManager<clock>::run() -> void
{
	std::uint64_t loop{0};
	std::vector<std::uint64_t> doneTimerIds;
	wzlogs_d("Running TimerManager in thread: " << std::this_thread::get_id());
	stopped_ = false;
	while (!shouldStop_)
	{
		loop++;
		auto now = clock::now();
		wzlogs_i(loop << ".Running timer handler ");
		auto nextTime = clock::now() + IDLE_TIMER_INTERVAL;

		{  // block in which the timers_ map is protected by a mutex
			std::unique_lock<std::mutex> lk(timer_mut_);
			wzlogs_i("Currently scheduled timers: " << timers_.size());
			doneTimerIds.clear();
			// find all the timers that are ready to run
			for (auto it = timers_.begin(); it != timers_.end();
			     /* no increment */)
			{
				if (it->first <= now)
				{
					std::ostringstream oss;
					oss << "Round:" << loop << ": Timer with id " << it->second.id << " is done, milliseconds too late:"
					    << std::chrono::duration_cast<std::chrono::milliseconds>(now - it->first).count();
					wzlog_i(wzlogger, oss.str());
					doneTimerIds.push_back(it->second.id);
					lk.unlock();
					if (it->second.callback)
					{
						it->second.callback();
					}
					if (shouldStop_) break;
					lk.lock();
					it = timers_.erase(it);
				}
				else
				{
					++it;
				}
			}  // end for timers
		}
		// Invoke callback for all the timers that were ready
		if (doneTimerIds.size() > 0)
		{
			std::lock_guard<std::mutex> lk(cb_mut_);
			if (multiCallback_)
			{
				multiCallback_(doneTimerIds);
			}
		}

		if (!shouldStop_)
		{  // block in which the timers_ map is protected by a mutex
			std::unique_lock<std::mutex> lk(timer_mut_);
			if (timers_.size() > 0)
			{
				nextTime = timers_.begin()->first;
			}
		}
		if (!shouldStop_)
		{
			std::unique_lock<std::mutex> cvlk(cv_mut_);
			reschedule_ = false;
			wzlogs_d("Going to sleep until " << toString(nextTime));
			auto dur = std::chrono::duration_cast<std::chrono::milliseconds>(nextTime - now);
			wzlogs_d("Sleeping (ms):  " << dur.count() << " ");

			auto status = cv_.wait_until(cvlk, nextTime, [this]() { return reschedule_; });
			if (status)
			{
				wzlogs_d("Wake up because of reschedule");
			}
			else
			{
				wzlogs_d("Wake up because of timeout");
			}
		}
	}
	stopped_ = true;
	cvFinished_.notify_all();
	wzlogs_d("TimerManager run function in thread: " << std::this_thread::get_id() << " done");
}

template<class clock>
auto TimerManager<clock>::getNextFreeTimerId() -> std::uint64_t
{
	std::lock_guard<std::mutex> lk(timer_mut_);
	const auto result = idPool_.front();
	idPool_.pop_front();
	return result;
}

template<class clock>
auto TimerManager<clock>::init() -> void
{
	wzlogs_d("Initialize timer service");
	{
		std::lock_guard<std::mutex> lk(cv_mut_);
		shouldStop_ = false;
	}
}

template<class clock>
auto TimerManager<clock>::shutdown() -> void
{
	wzlogs_d("Shuting down timer service");
	{
		std::lock_guard<std::mutex> lk(cv_mut_);
		shouldStop_ = true;
	}
	notifyReschedule();
}

template<class clock>
auto TimerManager<clock>::waitForFinished() -> void
{

	wzlogs_d("Waiting for timermanager run function to complete");
	UniqueLock lk{cvFinished_mut_};
	cvFinished_.wait(lk, [this]() { return stopped_.load(); });
}

template<class clock>
auto TimerManager<clock>::setTimersCallback(AllTimersCallback callback) -> void
{
	std::lock_guard<std::mutex> lk(cb_mut_);
	multiCallback_ = callback;
}

}  // namespace wzcpp

#endif  // __WZCPP_TIMER_H__