//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_UTILS_FORMATTER_H__
#define __WZCPP_UTILS_FORMATTER_H__


#include "String.h"
#include "StringView.h"
#include <iomanip>
#include <sstream>


////////////////////////////////////////////////////////////////////////////////
//
// 	 Include fmt which is used for inline std::string wzlf(...)
//   If Spdlog is used the bundled version is included otherwise
//   include the standalone header only version.
//
#ifdef SPDLOG_FMT_EXTERNAL
#include "spdlog/fmt/fmt.h"
#else
#ifndef FMT_HEADER_ONLY
#define FMT_HEADER_ONLY
#endif
#include "fmt/format.h"
#endif
////////////////////////////////////////////////////////////////////////////////

namespace wzcpp
{
/*!
    Takes a string with custom syntax, copies the string to out and
    replaces the syntax with fmt equivalents.

    @param in   The string with the custom syntax
    @param out  An empty string where the fmt syntax is written to
*/
inline void preformat(const string_view& in, std::string& out)
{
	out.insert(0, in.data(), in.size());

	// Replace hex syntax
	strReplaceFirst(out, "{%hex2}", "0x{:02X}");
	strReplaceFirst(out, "{%hex4}", "0x{:04X}");
	strReplaceFirst(out, "{%hex8}", "0x{:08X}");

	// Replace bin syntax
	strReplaceFirst(out, "{%bin4}", "0b{:04b}");
	strReplaceFirst(out, "{%bin8}", "0b{:08b}");

	// Replace Poco format
	strReplaceAll(out, "%?i", "{}");    // from bool to unsigned (hex)
	strReplaceAll(out, "%b", "{}");     // bool
	strReplaceAll(out, "%c", "{:c}");   // char
	strReplaceAll(out, "%d", "{:d}");   // int
	strReplaceAll(out, "%i", "{:d}");   // int
	strReplaceAll(out, "%u", "{:d}");   // unsigned
	strReplaceAll(out, "%hd", "{:d}");  // short
	strReplaceAll(out, "%hu", "{:d}");  // unsigned short
	strReplaceAll(out, "%ld", "{:d}");  // long
	strReplaceAll(out, "%lu", "{:d}");  // unsigned long
	strReplaceAll(out, "%Ld", "{:d}");  // long long
	strReplaceAll(out, "%Lu", "{:d}");  // unsigned long long
	strReplaceAll(out, "%o", "{:o}");   // unsigned (octal)
	strReplaceAll(out, "%x", "{:x}");   // unsigned (hex)
	strReplaceAll(out, "%X", "{:X}");   // unsigned (hex)
	strReplaceAll(out, "%f", "{:f}");   // double
	strReplaceAll(out, "%e", "{:e}");   // double
	strReplaceAll(out, "%E", "{:E}");   // double
	strReplaceAll(out, "%hf", "{:f}");  // float
	strReplaceAll(out, "%he", "{:e}");  // float
	strReplaceAll(out, "%hE", "{:E}");  // float
	strReplaceAll(out, "%s", "{:s}");   // std::string
	strReplaceAll(out, "%z", "{:d}");   // std::size_t
}


/*!
    Formats a string with a variadic set of arguments.
    Calls internally the fmt library function ::format().

    @param fmtStr   A string containing placeholders and formatting syntax
                    see https://fmt.dev/latest/syntax.html for the syntax
    @param args     A variadic parameter pack containing the arguments
    @returns        The formatted string

    @remarks        There exists a different very simple syntax for the basic
                    formatting use cases.
                    {%hex2} : 15 -> 0x0F
                    {%hex4} : 15 -> 0x000F
                    {%hex8} : 15 -> 0x0000000F
                    {%bin4} : 15 -> 0b1111
                    {%bin8} : 15 -> 0b00001111
*/
template<class... TArgs>
inline std::string wzlf(const std::string& fmtStr, TArgs&&... args)
{
	std::ostringstream stream;

	// Preformat the basic syntax in a fmt syntax
	std::string preformatString;
	preformat(fmtStr, preformatString);

	stream << fmt::format(preformatString, args...);
	return stream.str();
}

/*!
    Formats a byte container content to a string of format like "AF FE"

    @param v The byte container
*/
template<class Container>
inline std::string rangeToHex(Container v)
{
	std::ostringstream stream;

	auto first = begin(v);
	auto last  = end(v);
	if (first != last)
	{  // initial check
		while (true)
		{
			stream << std::uppercase << std::setw(2) << std::setfill('0') << std::hex
			       << static_cast<short>(*first++);  // Write to stream in format 0F
			if (first == last) break;
			stream << " ";
		}
	}

	return stream.str();
}

/*!
    Formats the value between start and end iterator into string of format like "AF FE"

    @param first Iterator to the first element of the container to format
    @param last Iterator to the last element of the container to format( The element to which last points is excluded)
*/
template<typename TIter>
inline std::string rangeToHex(TIter first, TIter last)
{
	std::ostringstream stream;
	if (first != last)
	{  // initial check
		while (true)
		{
			stream << std::uppercase << std::setw(2) << std::setfill('0') << std::hex
			       << static_cast<short>(*first++);  // Write to stream in format 0F
			if (first == last) break;
			stream << " ";
		}
	}
	return stream.str();
}
}  // namespace wzcpp
#endif  // !__WZCPP_UTILS_FORMATTER_H__
