//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_UTILS_STRING_H__
#define __WZCPP_UTILS_STRING_H__
#include <algorithm>
#include <cctype>
#include <cstring>
#include <exception>
#include <sstream>
#include <string>
#include <vector>

namespace wzcpp
{
/*!
    Replaces the first sequence of characters matching old with another sequence of
    characters.

    @param old      The sequence to replace
    @param fresh    The sequence old will be replaces for
    @returns        True if old could be found and was replaced
*/
inline bool strReplaceFirst(std::string& str, const char* old, const char* fresh)
{
	// Find start position
	size_t start_pos = str.find(old);

	// Return false if the end of the string is reached
	if (start_pos == std::string::npos)
	{
		return false;
	}

	// Replace old with fresh
	str.replace(start_pos, std::strlen(old), fresh);

	// All went well, return true
	return true;
}

/*!
    Replaces the all sequences of characters matching old with another sequence of
    characters.

    @param old      The sequence to replace
    @param fresh    The sequence old will be replaces for
    @returns        True if old could be found and was replaced
*/
inline void strReplaceAll(std::string& str, const char* old, const char* fresh)
{
	// Return if old is empty
	if (std::strlen(old) == 0)
	{
		return;
	}

	size_t start_pos = 0;
	while ((start_pos = str.find(old, start_pos)) != std::string::npos)
	{
		str.replace(start_pos, std::strlen(old), fresh);

		// In case 'fresh' contains 'old', like replacing 'x' with 'yx'
		start_pos += std::strlen(fresh);
	}
}

/*!
    Splits a string into tokens separated by delimiter
    @param s      the source string
    @param delimiter the character to split on
    @returns        vector of std::strings containing the parts
*/
inline std::vector<std::string> split(const std::string& s, char delimiter)
{
	std::vector<std::string> tokens;
	std::string token;
	std::stringstream tokenStream;
	tokenStream << s;
	while (std::getline(tokenStream, token, delimiter))
	{
		tokens.push_back(token);
	}
	return tokens;
}

/*!
    Converts a string into an unsigned int value
    @param s      the source string
    @returns      the integer parsed from the string
*/
inline unsigned int stoui(const std::string& s)
{
	if (s.find('-') != std::string::npos)  // Found - sign in string
	{
		throw std::out_of_range("Conversion of string to unsigned int failed. No negative numbers allowed");
	}
	unsigned long lresult = std::stoul(s, 0, 10);
	unsigned int result   = static_cast<unsigned int>(lresult);
	if (result != lresult)
	{
		throw std::out_of_range("Conversion of string to unsigned int failed");
	}
	return result;
}

/*!
    Converts a string into all lower case characters
    @param s      the source string
    @returns      a std::string with lowercase characters only
*/
inline auto toLower(const std::string& s) -> std::string
{
	std::string rs;
	rs.reserve(s.size());
	for (auto&& c : s)
	{
		rs += static_cast<char>(std::tolower(static_cast<unsigned char>(c)));
	}
	return rs;
}

/*!
    Converts a string into all upper case characters
    @param s      the source string
    @returns      a std::string with upper characters only
*/
inline auto toUpper(const std::string& s) -> std::string
{
	std::string rs;
	rs.reserve(s.size());
	for (auto&& c : s)
	{
		rs += static_cast<char>(std::toupper(static_cast<unsigned char>(c)));
	}
	return rs;
}

/*!
    Checks if a string contains only a decimal number
    @param numString	The string to test
    @returns true if string only starts with optional +- and contains only digits 0-9 afterwards, false otherwise
 */
inline auto isDecimal(const std::string& numString) -> bool
{
	int offset = 0;
	if (numString.front() == '+' || numString.front() == '-')
	{
		offset = 1;
	}
	return std::all_of(numString.begin() + offset, numString.end(), [](unsigned char c) { return std::isdigit(c); });
}

/*!
    Checks if a string contains only a hexadecimal number
    @param numString	The string to test
    @returns true if string starts with optional 0x or 0X and then only contains 0123456789abcdefABCDEF, false otherwise
 */
inline auto isHex(const std::string& numString) -> bool
{
	int offset = 0;
	if (numString.size() > 1 && toLower(numString.substr(0, 2)) == "0x")
	{
		offset = 2;
	}
	return std::all_of(numString.begin() + offset, numString.end(), [](unsigned char c) { return std::isxdigit(c); });
}

/*!
    Checks if a string contains a binary number
    @param numString	The string to test
    @returns	true if string start with 0b or 0B and then only 0s and 1s afterwards, false otherwise
 */
inline auto isBinary(const std::string& numString) -> bool
{
	int offset = 0;
	if (numString.size() > 1 && toLower(numString.substr(0, 2)) == "0b")
	{
		offset = 2;
	}
	return std::all_of(numString.begin() + offset, numString.end(), [&](char c) { return (c == '0' || c == '1'); });
	return false;
}

/*!
    Checks if a string contains a octal number
    @param numString	The string to test
    @returns	true if string starts with 0o or 0O and then only digits afterwards, false otherwise
 */
inline auto isOctal(const std::string& numString) -> bool
{
	int offset = 0;
	if (numString.size() > 1 && toLower(numString.substr(0, 2)) == "0o")
	{
		offset = 2;
	}
	return std::all_of(numString.begin() + offset, numString.end(), [](unsigned char c) { return std::isdigit(c); });
	return false;
}

/*!
    Returns true if numString is a number
    @param numString	The string to test
    @returns true is string cotains only a number, false otherwise
 */
inline auto isNumber(const std::string& numString) -> bool
{
	return (isDecimal(numString) || isHex(numString) || isBinary(numString) || isOctal(numString));
}


}  // namespace wzcpp

#endif  // !__WZCPP_UTILS_STRING_H__
