//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __WZCPP_UTILS_VERSION_H__
#define __WZCPP_UTILS_VERSION_H__
#include <cstring>
#include <sstream>
#include <string>
#include "String.h"


namespace wzcpp
{

/*!
    A class containing an application version in the format major, minor, patch
*/
class Version
{
public:
	Version() : major_(0), minor_(0), patch_(0)
	{
	}
	~Version() = default;
	Version(unsigned int major, unsigned int minor, unsigned int patch) : major_(major), minor_(minor), patch_(patch)
	{
	}

	Version(const std::string& str, const unsigned char delim = '.') : major_(0), minor_(0), patch_(0)
	{
		fromString(str, delim);
	}

	unsigned int getMajor() const
	{
		return major_;
	};

	unsigned int getMinor() const
	{
		return minor_;
	};

	unsigned int getPatch() const
	{
		return patch_;
	};

	/*!
	Sets all three parts at once
	@param major  the major version
	@param minor  the minor version
	@param patch  the patch version
	*/
	void setVersion(unsigned int major, unsigned int minor, unsigned int patch)
	{
		major_ = major;
		minor_ = minor;
		patch_ = patch;
	};


	/*!
	Splits the give string into substrings separated by the delim parameter, and then converts each part into an
	unsigned integer
	@param str  the string from witch the version should be parsed
	@param delim  the delimeter separating the individual parts inside the version string
	*/
	void fromString(const std::string& str, char delim = '.')
	{
		auto tokens = split(str, delim);
		if (tokens.size() != 3)
		{
			throw std::logic_error("Version string must contain major, minor and patch part");
		}
		major_ = stoui(tokens[0]);
		minor_ = stoui(tokens[1]);
		patch_ = stoui(tokens[2]);
	}

	std::string toString(char delim = '.') const
	{
		std::stringstream ss;
		ss << major_ << delim << minor_ << delim << patch_;
		return ss.str();
	}

	bool empty() const
	{
		return (major_ == 0) && (minor_ == 0) && (patch_ == 0);
	}

	friend bool operator==(const Version& lhs, const Version& rhs)
	{
		return ((lhs.major_ == rhs.major_) && (lhs.minor_ == rhs.minor_) && (lhs.patch_ == rhs.patch_));
	};

	friend bool operator!=(const Version& lhs, const Version& rhs)
	{
		return !((lhs.major_ == rhs.major_) && (lhs.minor_ == rhs.minor_) && (lhs.patch_ == rhs.patch_));
	};

	friend bool operator>(const Version& lhs, const Version& rhs)
	{
		return ((lhs.major_ > rhs.major_) || ((lhs.major_ == rhs.major_) && (lhs.minor_ > rhs.minor_)) ||
		        ((lhs.major_ == rhs.major_) && (lhs.minor_ == rhs.minor_) && (lhs.patch_ > rhs.patch_)));
	};

	friend bool operator<(const Version& lhs, const Version& rhs)
	{
		return ((lhs.major_ < rhs.major_) || ((lhs.major_ == rhs.major_) && (lhs.minor_ < rhs.minor_)) ||
		        ((lhs.major_ == rhs.major_) && (lhs.minor_ == rhs.minor_) && (lhs.patch_ < rhs.patch_)));
	};

	friend bool operator<=(const Version& lhs, const Version& rhs)
	{
		return (lhs < rhs) || (lhs == rhs);
	};

	friend bool operator>=(const Version& lhs, const Version& rhs)
	{
		return (lhs > rhs) || (lhs == rhs);
	};

private:
	unsigned int major_, minor_, patch_;
};

/*!
    A class containing an application version range.
    The boundaries are always included.
*/
class VersionRange
{
public:
	VersionRange()  = default;
	~VersionRange() = default;
	VersionRange(const Version& startVersion, const Version& endVersion)
	{
		startVersion_ = startVersion;
		endVersion_   = endVersion;
	}

	VersionRange(const Version& startVersion)
	{
		startVersion_ = startVersion;
	}

	Version getEndVersion() const
	{
		return endVersion_;
	}

	Version getStartVersion() const
	{
		return startVersion_;
	}

	bool contains(const Version& version) const
	{
		if (endVersion_.empty() && !startVersion_.empty())
		{
			return version >= startVersion_;
		}
		if (!endVersion_.empty() && !startVersion_.empty())
		{
			return (version >= startVersion_) && (version <= endVersion_);
		}
		if (!endVersion_.empty() && startVersion_.empty())
		{
			return (version <= endVersion_);
		}
		return false;
	}

private:
	Version startVersion_, endVersion_;
};


}  // namespace wzcpp

#endif  // !__WZCPP_UTILS_VERSION_H__
