//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "baos/BaosLib_export.h"

#include "wzcpp/features/error/Exception.h"

#ifndef __BAOSLIB_EXCEPTION_H__
#define __BAOSLIB_EXCEPTION_H__

/// @cond DOCUMENT_STL
#include <sstream>
/// @endcond

/*!
    @file
    @brief Custom exceptions for the SDK
*/

namespace wz::baos
{


/*!
    @brief Custom error codes for each exception

    Each custom exception has it's own error code. The reasoning behind this is that when
    creating wrapper around this code for other languages like C, that do not support
    exceptions its easy to catch the SDK exceptions extract the errorcode and forward it.
 */
enum class ErrorCodes
{
	BaosError             = 2,
	SocketError           = 3,
	AddressResolveError   = 4,
	TCPConnectionError    = 5,
	TCPConnectionClosed   = 6,
	SocketRecvError       = 7,
	SocketSendError       = 8,
	SocketInitError       = 9,
	SocketRxBufferToSmall = 10,
	ResponseError         = 20,
	ResponseTimeoutError  = 21,
	MaleformedTelegram    = 22,
	ClosedConnection      = 51,
	StoppedConnection     = 52
};

DECLARE_WZ_EXCEPTION(WzBaosException,
                     ErrorCodes::BaosError,
                     wzcpp::error::WzBaseException);  //!< Generic exception from this SDK
DECLARE_WZ_EXCEPTION(WzSocketException,
                     ErrorCodes::SocketError,
                     WzBaosException);  //!< Thrown if errors occur in the underlying sockets
DECLARE_WZ_EXCEPTION(WzIpAddressResolveException,
                     ErrorCodes::AddressResolveError,
                     WzSocketException);  //!< Thrown if the ip address could not be resolved
DECLARE_WZ_EXCEPTION(WzTCPConnectionException,
                     ErrorCodes::TCPConnectionError,
                     WzSocketException);  //!< Thrown for error during establishing of the TCP connection
DECLARE_WZ_EXCEPTION(WzTCPConnectionClosedException,
                     ErrorCodes::TCPConnectionClosed,
                     WzSocketException);  //!< Thrown if the TCP connection is unexpectely closed
DECLARE_WZ_EXCEPTION(WzRecvException,
                     ErrorCodes::SocketRecvError,
                     WzSocketException);  //!< Thrown on error during a read on a socket
DECLARE_WZ_EXCEPTION(WzSendException,
                     ErrorCodes::SocketSendError,
                     WzSocketException);  //!< Thrown on error during a send to a socket
DECLARE_WZ_EXCEPTION(WzSocketInitException,
                     ErrorCodes::SocketInitError,
                     WzSocketException);  //!< Thrown if creation or initialization of the socket failed
DECLARE_WZ_EXCEPTION(WzRxBufferOverflowException,
                     ErrorCodes::SocketRxBufferToSmall,
                     WzSocketException);  //!< Thrown if the incoming telegram is to big for the buffer. The buffer size
                                          //!< is defined in MAX_TELEGRAM_LENGTH

DECLARE_WZ_EXCEPTION(WzMaleformedTelegramException,
                     ErrorCodes::MaleformedTelegram,
                     WzBaosException);  //!< Thrown if incoming telegrams are not wellformed
DECLARE_WZ_EXCEPTION(WzResponseException,
                     ErrorCodes::ResponseError,
                     WzBaosException);  //!< Thrown if the response contains an errorcode other then success
DECLARE_WZ_EXCEPTION(WzResponseTimeoutException,
                     ErrorCodes::ResponseTimeoutError,
                     WzBaosException);  //!< Thrown if an expected response was not received in time
DECLARE_WZ_EXCEPTION(WzClosedConnectionException,
                     ErrorCodes::ClosedConnection,
                     WzBaosException);  //!< Thrown if the connection was not open for some task
DECLARE_WZ_EXCEPTION(
    WzStoppedConnectionException,
    ErrorCodes::StoppedConnection,
    WzBaosException);  //!< Thrown if attempted to open a connection which underlying system is not running

}  // namespace wz::baos

#endif  // __BAOSLIB_EXCEPTION_H__
