//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_CONNECTION_H__
#define __BAOSLIB_CONNECTION_H__

#include "baos/config.h"
#include "baos/connection/IOEvent.h"
#include "baos/ip/IpGeneral.h"
#include "baos/ip/TcpIO.h"
#include "baos/protocol/Defines.h"
#include "baos/protocol/Response.h"
#include "connection/BaosIndicationSender.h"
#include "ip/TcpDriver.h"

#include "wzcpp/features/concurrent/executor.h"
#include "wzcpp/features/logging.h"
#include "wzcpp/features/timer/timer.h"

/// @cond DOCUMENT_STL
#include <future>
#include <map>
#include <memory>
#include <mutex>
#include <optional>
#include <string>
/// @endcond

/*!
    @file
    @brief Code regarding a connection to a physical BAOS device
*/

namespace wz::baos
{

/// The different result type of the BAOS binary parsing operation
enum class BaosParseStatus
{
	Success,
	SizeMismatch,
	InvalidKnxHeader,
	InvalidConnectionHeader,
	InvalidMsgData,
	InvalidMainService,
	InvalidSubService,
	UnknownIndicationType,
	MalformedResponse
};

std::string BAOSLIB_EXPORT
parseStatusToString(BaosParseStatus& parseStatus);  //!< Human readable string of the parse status.

/*!
    @brief Represents a connection to a BAOS device

    The connection runs in a separate thread and receives events from either the user or from underlying socket.
 */
class BAOSLIB_EXPORT BaosConnection
{
public:
	using Clock          = std::chrono::system_clock;  //!< An  alias for the clock internaly used.
	using BaosFrame      = Buffer;  //!< An alias for the container type to store frame/packets/telegrams in.
	using BaosFrame_UPtr = std::unique_ptr<BaosFrame>;       //!< An alias for a std::unique_ptr holding a BaosFrame.
	using UPtr           = std::unique_ptr<BaosConnection>;  //!< An alias for an unqiue pointer to this class
	using ServerItemIndicationCb =
	    std::function<void(const protocol::ServerItems&)>;  //!< Function signature for serveritem indication callbacks
	using DatapointValueCb = std::function<void(
	    const protocol::DatapointValueStates&)>;  //!< Function signature for datapoint value indication callbacks

	BaosConnection();           //!< Default constructor
	virtual ~BaosConnection();  //!< Default destructor

	void start();  //!< Start the client side of the connection
	void stop();   //!< Stop the client side of the connection

	virtual void reopen() = 0;  //!< Reopens the underlying connection, has to be implemented by derived classes
	void close();               //!< Close the connection to the BAOS device

	/*!
	    @brief Returns the open state of the connection

	    @return true if the connection is open, false otherwise
	 */
	bool isOpen();


	/*!
	    @name Synchronous API
	    Set of functions to provide the typical BAOS interaction in a synchronous way
	 */
	///@{

	/*!
	@brief Get a single the ServerItem

	@param id  The ServerItem id
	@return The protocol::ServerItem value or an empty std::optional if no value this ServerItemId could be retrieved
	 */
	std::optional<protocol::ServerItem> getServerItem(protocol::ServerItemId id); /*!< */

	/*!
	    @brief Get a list of serveritems from the device

	    @param startItem The id of the first serveritem to load
	    @param numberOfItems Maximal number of items to return
	    @return protocol::ServerItems a list of server items in the interval [startItem … startItem+numberOfItems-1]
	 */
	protocol::ServerItems getServerItems(std::uint16_t startItem, std::uint16_t numberOfItems);

	void setServerItem(const protocol::ServerItem& serverItem); /*!< Set a single serveritem */
	void setServerItem(protocol::ServerItem&& serverItem);      /*!< Set a single serveritem */
	void setServerItems(protocol::ServerItems&& serverItems);   /*!< Set a list of serveritems */

	/*!
	    @brief Get the datapoint description for one datapoint

	    @param dpId  The 1 based id of the datapoint
	    @return The protocol::DatapointDescription value or an empty std::optional if no DatapointDescription for this
	   dpId could be retrieved
	 */
	std::optional<protocol::DatapointDescription> getDpDescription(std::uint16_t dpId);

	/*!
	    @brief Get a list of DatapointDescriptions

	    @param startDpId  The 1 based id of the first datapoint
	    @param numberOfItems Maximal number of descriptions to return
	    @return A map with datapoint id as key and  DatapointDescription as value  in the interval [startDpId …
	    startDpId + numberOfItems - 1]
	 */
	protocol::DatapointDescriptions getDpDescriptions(std::uint16_t startDpId, std::uint16_t numberOfItems);

	/*!
	    @brief Get the datapoint description string for one datapoint

	    @param dpId The 1 based id of the datapoint
	    @return The protocol::DatapointDescriptionString value or an empty std::optional if no
	   DatapointDescriptionString for this dpId could be retrieved
	 */
	std::optional<protocol::DatapointDescriptionString> getDpDescriptionString(std::uint16_t dpId);

	/*!
	    @brief Get a list of DatapointDescriptionsStrings

	    @param startDpId The 1 based id of the first datapoint
	    @param numberOfStrings Maximal number of descriptions to return
	    @return A map with datapoint id as key and DatapointDescriptionsString as value in the interval [startDpId …
	    startDpId + numberOfItems - 1]
	 */
	protocol::DatapointDescriptionStrings getDpDescriptionStrings(std::uint16_t startDpId,
	                                                              std::uint16_t numberOfStrings);

	std::optional<std::uint8_t> getParameterByte(std::uint16_t byteIndex);  //!< Get a single parameter byte

	/*!
	    @brief Get a list of parameter bytes

	    @param byteIndex The 1 based index of the first parameter byte to read
	    @param numberOfBytes Maximum number of bytes to return
	    @return protocol::ParameterBytes a list of parameter bytes in the interval  [byteIndex …
	   byteIndex+numberOfBytes-1]
	 */
	protocol::ParameterBytes getParameterBytes(std::uint16_t byteIndex, std::uint16_t numberOfBytes);

	std::optional<protocol::DatapointValueState> getDpValue(
	    std::uint16_t dpId);  //!< Get a single datapoint value with its associated state

	/*!
	    @brief Get a map with the datapoint id as key and the DatapointValueState as value

	    @param startDpId The 1 based id of the first datapoint
	    @param numberOfValues Maximal number of value to return
	    @param filter [optional] Filter for the requested datapoints
	    @return A map with datapoint id as key and DatapointValueState as value in the interval
	    [startDpId...startDpId+numberOfValues-1]
	 */
	protocol::DatapointValueStates getDpValues(
	    std::uint16_t startDpId,
	    std::uint16_t numberOfValues,
	    protocol::DatapointValueFilter filter = protocol::DatapointValueFilter::ALL);

	/*!
	    @brief Send a SetDatapointValue request to the BAOS device for one datapoint

	    @param dpId  The 1 based id of the datapoint
	    @param dpValue The new value to set and the command for it
	 */
	void setDpValue(std::uint16_t dpId, protocol::DatapointChangeValue&& dpValue);


	/*!
	    @brief Send a SetDatapointValue request to the BAOS device with given datapoint values

	    @param dpValues a map with datapoint ids as keys and command and new value as value
	 */
	void setDpValues(std::map<std::uint16_t, protocol::DatapointChangeValue>&& dpValues);
	///@}

	/******************************************************************************************************************/
	/*!
	    @name Asynchronous API
	    Set of functions to provide the typical BAOS interaction in a asynchronous way. Every function returns a future.
	 */
	///@{
	protocol::ResponseFuture getServerItemAsync(
	    protocol::ServerItemId id); /*!< Get a single serveritem specified by its id from the device */

	/*!
	    @brief Get a list of serveritems from the device

	    @param startItem The id of the first serveritem to load
	    @param numberOfItems Maximal number of items to return
	    @return A ReponseFuture with the payload containing a list of server items in the interval [startItem …
	    startItem+numberOfItems-1]
	 */
	protocol::ResponseFuture getServerItemsAsync(std::uint16_t startItem, std::uint16_t numberOfItems);

	protocol::ResponseFuture setServerItemAsync(protocol::ServerItem&& serverItem);    /*!< Set a single serveritem */
	protocol::ResponseFuture setServerItemsAsync(protocol::ServerItems&& serverItems); /*!< Set a list of serveritems */

	/*!
	    @brief Get the datapoint description for one datapoint

	    @param dpId  The 1 based id of the datapoint
	    @return A ReponseFuture with the payload containing the protocol::DatapointDescription
	 */
	protocol::ResponseFuture getDpDescriptionAsync(std::uint16_t dpId);

	/*!
	    @brief Get a list of DatapointDescriptions

	    @param startDpId  The 1 based id of the first datapoint
	    @param numberOfItems Maximal number of descriptions to return
	    @return A ReponseFuture with the payload containing a map with datapoint id as key and  DatapointDescription as
	    value  in the interval [startDpId … startDpId + numberOfItems - 1]
	 */
	protocol::ResponseFuture getDpDescriptionsAsync(std::uint16_t startDpId, std::uint16_t numberOfItems);

	/*!
	    @brief Get the datapoint description string for one datapoint

	    @param dpId The 1 based id of the datapoint
	    @return A ReponseFuture with the payload containing the protocol::DatapointDescriptionString
	 */
	protocol::ResponseFuture getDpDescriptionStringAsync(std::uint16_t dpId);

	/*!
	    @brief Get a list of DatapointDescriptionsStrings

	    @param startDpId The 1 based id of the first datapoint
	    @param numberOfStrings Maximal number of descriptions to return
	    @return A ReponseFuture with the payload containing a map with datapoint id as key and
	   DatapointDescriptionsString as value in the interval [startDpId … startDpId + numberOfItems - 1]
	 */
	protocol::ResponseFuture getDpDescriptionStringsAsync(std::uint16_t startDpId, std::uint16_t numberOfStrings);

	protocol::ResponseFuture getParameterByteAsync(std::uint16_t byteIndex);  //!< Get a single parameter byte

	/*!
	    @brief Get a list of parameter bytes

	    @param byteIndex The 1 based index of the first parameter byte to read
	    @param numberOfBytes Maximum number of bytes to return
	    @return A ReponseFuture with the payload containing a list of parameter bytes in the interval  [byteIndex …
	    byteIndex+numberOfBytes-1]
	 */
	protocol::ResponseFuture getParameterBytesAsync(std::uint16_t byteIndex, std::uint16_t numberOfBytes);

	protocol::ResponseFuture getDpValueAsync(
	    std::uint16_t dpId);  //!< Get a single datapoint value with its associated state

	/*!
	    @brief Get a map with the datapoint id as key and the DatapointValueState as value

	    @param startDpId The 1 based id of the first datapoint
	    @param numberOfValues Maximal number of value to return
	    @param filter [optional] Filter for the requested datapoints
	    @return A ReponseFuture with the payload containing a map with datapoint id as key and DatapointValueState as
	   value in the interval [startDpId...startDpId+numberOfValues-1]
	 */
	protocol::ResponseFuture getDpValuesAsync(
	    std::uint16_t startDpId,
	    std::uint16_t numberOfValues,
	    protocol::DatapointValueFilter filter = protocol::DatapointValueFilter::ALL);


	/*!
	    @brief Send a SetDatapointValue request to the BAOS device for one datapoint

	    @param dpId  The 1 based id of the datapoint
	    @param dpValue The new value to set and the command for it
	    @return A ResponseFuture with no payload
	 */
	protocol::ResponseFuture setDpValueAsync(std::uint16_t dpId, protocol::DatapointChangeValue&& dpValue);

	/*!
	    @brief Send a SetDatapointValue request to the BAOS device with given datapoint values

	    @param dpValues a map with datapoint ids as keys and command and new value as value
	    @return A ResponseFuture with no payload
	 */
	protocol::ResponseFuture setDpValuesAsync(std::map<std::uint16_t, protocol::DatapointChangeValue>&& dpValues);

	/*!
	    @brief Close the connection

	    @return protocol::ResponseFuture  which completes if the connection is acutally closed
	 */
	protocol::ResponseFuture closeAsync();


	///@}

	/*!
	    Sets the function callback to be invoked in every serveritem indication from the device
	    @param callback  strong typed std::function object
	*/
	void setServerItemsIndCallback(ServerItemIndicationCb callback)
	{
		indSender_.setServerItemsIndCallback(callback);
	}

	/*!
	    Sets the function callback to be invoked in every datapoint value indication from the device
	    @param callback  strong typed std::function object
	*/
	void setDatapointValueIndCallback(DatapointValueCb callback)
	{
		indSender_.setDatapointValueIndCallback(callback);
	}

protected:
	/*!
	    @brief Sends a generic frame( a byte list) to the BAOS device

	    @param frame a generic list of bytes
	    @return protocol::ResponseFuture  completes if the request got its matching response from the device
	 */
	protocol::ResponseFuture send(BaosFrame&& frame)
	{
		return sendImpl(std::move(frame));
	};

protected:
	/*!
	    @brief Represents the internal state of this connection
	 */
	enum class State
	{
		RUNNING,       //!< The worker is started but no connection to the device is established
		STOPPING,      //!< A stop was requested but it is not done with stopping
		STOPPED,       //!< It is now actualy stopped
		SENDING,       //!< It is in the process of sending out bytes
		SENDCOMPLETE,  //!< A complete packet was send
		WAITRESPONSE,  //!< Waiting for a response
		IDLE,          //!< Nothing to do but connection to device is established and worker is started
		CONERROR,      //!< An error occured during communication with the remote device
		CONCLOSED      //!< The underlying connection is actually closed
	};

	//! The asynchronous functions return a promise to keep internal track of their state this enum is used
	enum class PromiseState
	{
		UNUSED,
		SET,
		RESOLVED
	};

	/*!
	    @brief Get human readable string from internal state
	 */
	static std::string stateToString(State state);  //!<

	/*!
	    @brief For every received BAOS frame this handler is internally called

	    @param frame an rvalue of the frame to handle
	    @return BaosParseStatus
	*/
	BaosParseStatus handleBaosFrame(BaosFrame_UPtr&& frame);

	/*!
	    @brief For every received BAOS indication this handler is internally called

	    @param frame an rvalue of the frame to handle
	    @return BaosParseStatus
	 */
	BaosParseStatus handleBaosIndication(BaosFrame_UPtr&& frame);

	/*!
	    @brief For every received BAOS response this handler is internally called

	    @param frame an rvalue of the frame to handle
	    @return BaosParseStatus
	 */
	protocol::BaosResponseCode handleBaosResponse(BaosFrame_UPtr&& frame);

	/*!
	    @brief Set the internal state of this connections state machine

	    @param newState The new state to set
	    @return State The old state
	 */
	State setState(State newState);

	/*!
	    @brief Get the current internal state of this connections state machine

	    @return State The current state
	 */
	State getState() const
	{
		std::scoped_lock<std::mutex> lock{stateMutex_};
		return state_;
	};

	/*!
	    @brief Decodes the baos response

	    @param frame an rvalue of the frame to decode
	    @return protocol::BaosBaseResponse::SPtr a shared pointer with the ResponseObject.
	 */
	protocol::BaosBaseResponse::SPtr decodeResponse(BaosFrame_UPtr&& frame);

	virtual protocol::ResponseFuture closeImplAsync();  //!< The implementation function for closing a connection
	virtual void startImpl();  //!< The implementation function for starting the connection class
	virtual void stopImpl();   //!< The implementation function for stopping the connection class

	/*!
	    @brief The implementation function for getting the opened state of the connection

	    @return Always false in the base class. Derived classes should overwrite this
	 */
	virtual bool isOpenImpl()
	{
		return false;
	};
	virtual protocol::ResponseFuture sendImpl(
	    BaosFrame&& frame) = 0;  //!< The implementation function for sending bytes to the connected device

	virtual void onStateChange(State oldState,
	                           State newState);  //!< Invoked if the internal connection state should be changed.
	virtual void onTimerExpired(std::vector<std::uint64_t> ids);  //!< Invoked if on of the timers has expired

protected:
	std::atomic<bool> shouldStop_{false};  //!< Flag to indicate that the connection should stop at the next opportunity

	/*!
	    @brief Keep track if the connection is in running state or stopping / stopped

	    It is also used be start and stop function to protect against starting a running or stopping a stopped
	   connection.
	 */
	std::atomic<bool> isRunning_{false};
	wzcpp::ThreadPoolExecutor executor_{4};  //!< The main executor thread pool with 4 threads : The read thread and the
	                                         //!< indication thread, the connection thread and the timer thread

	wzcpp::TimerManager<Clock> timerManager_;  //!< Manages the timers for this connection
	wzcpp::Timer<Clock> respTimeoutTimer_;     //!< The actual timer objects for respone timeouts
	wzcpp::Timer<Clock> heartBeatTimer_;       //!< The actual timer objects for heartbeats

	Buffer waitResponse_;  //!< While waiting for a response this holds the pattern of the response we expect.
	protocol::ResponsePromise
	    responsePromise_;  //!< If a packet is send out via handleTxTcp the corresponding promise is stored here
	PromiseState responsePromiseState_{PromiseState::UNUSED};  //!< The state of the responsePromise_
	std::uint64_t currentMessageId_{0};                        //!< For internal use a unique message id
	std::queue<IOEvent> txQueue_;                              //!< Holds the packets to send waiting to be processed
	BaosMsgQueue bMsgQ_;                                       //!< Hold the event for the main connection loop
	protocol::BaosIndQueue bIndQ_;                             //!< Holds the indication to be send out to clients
	BaosIndicationSender indSender_;                           //!< Responsible for sending out the indications

private:
	WZLOGGER("BaosConnection", LVL_DEBUG);
	State state_{State::STOPPED};  //!< The current state of the connection
	mutable std::mutex stateMutex_;
};

}  // namespace wz::baos

#endif  // __BAOSLIB_CONNECTION_H__
