//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_EVENT_H__
#define __BAOSLIB_EVENT_H__

#include "baos/BaosLib_export.h"
#include "baos/config.h"
#include "baos/ip/IpGeneral.h"
#include "baos/protocol/Response.h"
#include "wzcpp/features/concurrent/concurrent_queue.h"

/// @cond DOCUMENT_STL
#include <future>
#include <memory>
#include <variant>
/// @endcond

/*!
    @file
    @brief Code regarding Input/Output Events
*/

namespace wz::baos
{

/*!
    @brief Represents an IOEvent.

    @see wz::baos::BaosConnection
 */
class BAOSLIB_EXPORT IOEvent
{
public:
	using UPtr          = std::unique_ptr<IOEvent>;  //!< An alias for an unique pointer to an IOEvent
	using EventDataType = std::variant<ip::Ip4SocketAddress, std::unique_ptr<Buffer>>;  //!< The different possible
	                                                                                    //!< datatypes inside an IOEvent

	//! The different possible event types
	enum class Type
	{
		UNKNOWN,          //!< If event is not specified
		OPEN,             //!< An open event is requested
		CLOSE,            //!< An close event is requested
		SOCKETERROR,      //!< Errors regarding the underlying socket
		RESPONSETIMEOUT,  //!< The response was not received in the through timeout given period
		HEARTBEAT,        //!< A Heartbeat is requested
		RX_PACKET,        //!< An incoming (RX) packet was received and should be handled
		TX_PACKET         //!< An outgoing (TX) packet should be handled
	};
	IOEvent() = default;  //!< Default constructor, generating an empty IOEvent

	/*!
	    @brief Construct a new IOEvent object

	    @param eventT The event type
	    @param pkt The data of this event as a std::variant
	    @param prom The promise which will be resolved if the event is handled
	    @param mID An unique message id to help with correlate the event
	 */
	IOEvent(Type eventT, EventDataType&& pkt, protocol::ResponsePromise&& prom, std::uint64_t mID = 0)
	    : eventType{eventT},
	      msgID{mID},
	      responsePromise{std::move(prom)}
	{
		data_ = std::move(pkt);
	}

public:
	Type eventType{Type::UNKNOWN};              //!< Stors the event type
	std::uint64_t msgID{0};                     //!< Stores the msgID
	protocol::ResponsePromise responsePromise;  //!< Stores the response promise to resolve later
	bool promiseMoved{false};  //!< Flag to indicate thate the response promise was moved and is therefore invalid
	EventDataType data_;       //!< The payload/data for this event
};

/*!
    @brief Provides a human readable string for an IOEvent type

    @param typ The IOEvent type
    @return std::string
 */
inline std::string toString(IOEvent::Type typ)
{
	switch (typ)
	{
		case IOEvent::Type::UNKNOWN:
			return "UNKNOWN";
		case IOEvent::Type::OPEN:
			return "OPEN";
		case IOEvent::Type::CLOSE:
			return "CLOSE";
		case IOEvent::Type::SOCKETERROR:
			return "SOCKETERROR";
		case IOEvent::Type::RESPONSETIMEOUT:
			return "RESPONSETIMEOUT";
		case IOEvent::Type::HEARTBEAT:
			return "HEARTBEAT";
		case IOEvent::Type::RX_PACKET:
			return "RX_PACKET";
		case IOEvent::Type::TX_PACKET:
			return "TX_PACKET";
		default:
			return "Unhandled IOEvent type";
	}
}

using BaosMsgQueue = wzcpp::ConcurrentDeque<IOEvent>;  //!< Alias for a concurrent queue with IOEvents

}  // namespace wz::baos

#endif  //__BAOSLIB_TCP_IO_H__
