//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_INTERFACE_DATA_TYPES_H__
#define __BAOSLIB_INTERFACE_DATA_TYPES_H__

#include "baos/config.h"

#include "BlockReader.h"
#include "baos/connection/BaosConnection.h"
#include "baos/device/BaosDevicesMeta.h"
#include "baos/protocol/Defines.h"

#include "wzcpp/features/algorithm/append.h"
#include "wzcpp/features/logging.h"
#include "wzcpp/features/utils/ByteArray.h"
#include "wzcpp/features/utils/Memory.h"

/// @cond DOCUMENT_STL
#include <memory>
#include <optional>
#include <utility>
/// @endcond

/*!
    @file
    @brief Specific datatypes used by the device classes
*/

namespace wz::baos
{

/*!
    @brief A collection of fields representing those serveritems which won't ever change in a device.

    Things like the serial number, the hardware type, will not change during normal operation of the device.
   Firmwarupdates on the other hand may as well change those fields.
 */
struct BAOSLIB_EXPORT DeviceStaticInfo
{
	wzcpp::SixByteArray hardwareType{0x0, 0x0, 0x0, 0x0, 0x0, 0x0};  //!< The hardware type of the device
	std::uint8_t hardwareVersion{0};                                 //!< The hardware version of the device
	std::uint8_t firmwareVersion{0};                                 //!< The firmware version of the device
	std::uint16_t manufactureCodeDevice{0};                          //!< The manufacturer code of the device
	wzcpp::SixByteArray serialNumber{0x0, 0x0, 0x0, 0x0, 0x0, 0x0};  //!< manufacturer specific serial number
	std::uint16_t maxBufferSize{0};              //!< The maximum packet length for BAOS binary packets supported
	std::uint8_t binaryProtocolVersion{0};       //!< The BAOS binary protocol version supported
	std::uint8_t webServicesProtocolVersion{0};  //!< The WEB services protocol version supported
	std::uint8_t restServiceProtocolVersion{0};  //!< The REST services protocol version supported
	wzcpp::SixByteArray macAddress{0x0, 0x0, 0x0, 0x0, 0x0, 0x0};  //!< The MAC address
	std::uint8_t maxManagementClients{0};                          //!< The maximum of parallel management connections
	std::uint8_t maxTunnellingClients{0};  //!< The maximum of parallel KNXnet/IP tunneling connections
	std::uint8_t maxBaosUdpClients{0};     //!< The maximum of parallel BAOS binary protocol UDP connections
	std::uint8_t maxBaosTcpClients{0};     //!< The maximum of parallel BAOS binary protocol TCP connections
	std::uint16_t maxDatapoints{0};        //!< The maximum amount of datapoint supported by this firmware
	std::uint16_t maxParameterBytes{0};    //!< The maximum amount of parameter bytes supported by this firmware

	static constexpr std::array ids{
	    protocol::ServerItemId::HardwareType,
	    protocol::ServerItemId::HardwareVersion,
	    protocol::ServerItemId::ManufactureCodeDevice,
	    protocol::ServerItemId::SerialNumber,
	    protocol::ServerItemId::MaxBufferSize,
	    protocol::ServerItemId::BinaryProtocolVersion,
	    protocol::ServerItemId::WebServicesProtocolVersion,
	    protocol::ServerItemId::RestServiceProtocolVersion,
	    protocol::ServerItemId::MacAddress,
	    protocol::ServerItemId::MaxManagementClients,
	    protocol::ServerItemId::MaxTunnellingClients,
	    protocol::ServerItemId::MaxBaosUdpClients,
	    protocol::ServerItemId::MaxBaosTcpClients,
	    protocol::ServerItemId::MaxDatapoints,
	    protocol::ServerItemId::MaxParameterBytes};  //!< A std::array with all the serveritem ids in this collection
};


/*!
    @brief A collection of fields representing those serveritems which may change if the configuration of the device is
    changed.

    Things like the individual address, the ip addres etc. may be changed during an ETS download or a configuration
   change by other means. But during the day to day operation of the device those are not changed.
 */
struct BAOSLIB_EXPORT DeviceConfigInfo
{
	std::uint16_t manufactureCodeApp{0};  //!< The manufacturer code of the ETS application e.g. 0x00C5 for Weinzierl
	std::uint16_t applicationId{0};      //!< ID of application loaded by ETS. It is mapped to bytes 2 and 3 of property
	                                     //!< PID_APPLICATION_VER in application object.
	std::uint8_t applicationVersion{0};  //!< Version of application loaded by ETS. It is mapped to byte 4 of property
	                                     //!< PID_APPLICATION_VER in application object.
	std::uint16_t lengthOfDescriptionString{0};  //!< The maximum length of a datapoint description string
	std::uint8_t baudrateItem{0};                //!< The current BAUD rate. Only useful for serial FT1.2 connections
	std::uint16_t currentBufferSize{0};          //!< The current maximum length for a BAOS telegram
	bool indicationSetting{
	    false};  //!< Is sending of indication via the BAOS binary protocol enabled(true) or disabled(false)
	std::uint16_t individualAddress{0};  //!< The KNX individual address
	bool tunnellingEnabled{false};       //!< Is the KNXnet/IP tunneling server enabled(true) or disabled(false)
	bool baosBinaryEnabled{false};       //!< Is the BAOS binary protocol enabled, This SDK relies on this to be true
	bool baosWebEnabled{false};          //!< Are the BAOS webservices on the device enabled(true) or disabled(false)
	bool baosRestEnabled{false};         //!< Are the BAOS REST services on the device enabled(true) or disabled(false)
	bool httpFileEnabled{false};         //!< Is the webserver on the device enabled(true) or disabled(false)
	bool searchRequestEnabled{false};    //!< Does the device react to a KNXnet/IP Search request(true) or not (false)
	bool isStructured{false};  //!< Some devices like the 777 may contain additional data about the building structure,
	                           //!< like rooms etc.
	std::uint16_t configuredDatapoints{0};  //!< How many datapoints are configured in the device
	std::string deviceFriendlyName;         //!< For IP devices the same as KNXnet/IP device friendly name
	std::uint16_t downloadCounter{0};       //!< The ETS download counter
	protocol::IpAssigmentMethod iPAssignment{
	    protocol::IpAssigmentMethod::dhcp};  //!< How the devices assigns its own IP address, In most cases the manual
	                                         //!< and dhco version are the viable options.
	std::array<std::uint8_t, 4> iPAddress;   //!< If manual IP assignment is selected, this is the ip address that the
	                                         //!< device should use for itself.
	std::array<std::uint8_t, 4> subnetMask;  //!< If manual IP assignment is selected, this is the subnet mask to set.
	std::array<std::uint8_t, 4>
	    defaultGateway;  //!< If manual IP assigment is selected, this is the default gateway to be set in the device.
	protocol::TimeSinceResetUnit timeSinceResetUnit;  //!< The unit in which timesincereset is represented
	std::int8_t SystemTimezoneOffset{0};              //!< The offset in hours of the system time to UTC
	bool menuEnabled{false};  //!< Are configuration changes via the device menu, visible on the display of the device
	                          //!< itself allowed or not
	bool suspendEnabled{
	    false};  //!< Is the suspend on idle mode ativated or not. Only usefull for USB devices at the moment.
	wzcpp::SixByteArray rfDomainAddress{
	    0x0, 0x0, 0x0, 0x0, 0x0, 0x0};  //!< The RF domain address. Only for device supported KNX RF as medium.
	protocol::StatusFlagsSupported statusFlagsSupported{
	    0x0000};  //!< A collection of flags with current device state and capabilities.
	static constexpr std::array ids{
	    protocol::ServerItemId::ManufactureCodeApp,
	    protocol::ServerItemId::ApplicationId,
	    protocol::ServerItemId::ApplicationVersion,
	    protocol::ServerItemId::LengthOfDescriptionString,
	    protocol::ServerItemId::BaudrateItem,
	    protocol::ServerItemId::CurrentBufferSize,
	    protocol::ServerItemId::IndicationSetting,
	    protocol::ServerItemId::IndividualAddress,
	    protocol::ServerItemId::TunnellingEnabled,
	    protocol::ServerItemId::BaosBinaryEnabled,
	    protocol::ServerItemId::BaosWebEnabled,
	    protocol::ServerItemId::BaosRestEnabled,
	    protocol::ServerItemId::HttpFileEnabled,
	    protocol::ServerItemId::SearchRequestEnabled,
	    protocol::ServerItemId::IsStructured,
	    protocol::ServerItemId::DeviceFriendlyName,
	    protocol::ServerItemId::ConfiguredDatapoints,
	    protocol::ServerItemId::DownloadCounter,
	    protocol::ServerItemId::IPAssignment,
	    protocol::ServerItemId::IPAddress,
	    protocol::ServerItemId::SubnetMask,
	    protocol::ServerItemId::DefaultGateway,
	    protocol::ServerItemId::TimeSinceResetUnit,
	    protocol::ServerItemId::SystemTimezoneOffset,
	    protocol::ServerItemId::MenuEnabled,
	    protocol::ServerItemId::SuspendEnabled,
	    protocol::ServerItemId::RFDomainAddress,
	    protocol::ServerItemId::StatusFlagsSupported};  //!< A std::array with all the serveritem ids in this collection
};


/*!
    @brief A collection of fields representing those serveritems which may change during "normal" operation of the
   device.

    Things like the programming mode, the bus connection state etc. may be changed during runtime.
 */
struct BAOSLIB_EXPORT DeviceRuntimeInfo
{
	std::uint32_t timeSinceReset{0};             //!< The time since the last reset of the device in TimeSinceResetUnits
	bool busConnected{false};                    //!< The TP KNX bus connection state
	bool programmingMode{false};                 //!< The KNX programming mode state
	std::uint8_t connectedManagementClients{0};  //!< How many management clients are currently connected
	std::uint8_t connectedTunnellingClients{0};  //!< How many KNXnet/IP tunneling clients are currently connected
	std::uint8_t connectedBaosUdpClients{0};     //!< How many BAOS clients are connected via UDP
	std::uint8_t connectedBaosTcpClients{0};     //!< How many BAOS clients are connected via TCP
	Buffer SystemTime;                           //!< The system time of the device in seconds since 1.1.1970 0:00
	protocol::StatusFlags statusFlags{
	    0x0000};  //!< The state of the status flags. @see wz::baos::protocol::StatusFlags for details
	static constexpr std::array ids{
	    protocol::ServerItemId::TimeSinceReset,
	    protocol::ServerItemId::BusConnected,
	    protocol::ServerItemId::ProgrammingMode,
	    protocol::ServerItemId::ConnectedManagementClients,
	    protocol::ServerItemId::ConnectedTunnellingClients,
	    protocol::ServerItemId::ConnectedBaosUdpClients,
	    protocol::ServerItemId::ConnectedBaosTcpClients,
	    protocol::ServerItemId::SystemTime,
	    protocol::ServerItemId::StatusFlags};  //!< A std::array with all the serveritem ids in this collection
};

/*!
    @brief Information about a datapoint object
 */
struct BAOSLIB_EXPORT DatapointConfig
{
	protocol::DatapointSize size;         /*!< The size as an enum value */
	protocol::DatapointConfigFlags flags; /*!< The knx group object flags*/
	protocol::DPT dpt;                    /*!< The datapoint type as on enum value */
	std::string desc;                     /*!< The description of the datapoint */
};

std::string BAOSLIB_EXPORT toString(const DatapointConfig& config);  //!< Convert to human readable string


using DatapointConfigurations = std::map<std::uint16_t, DatapointConfig>;  //!< A map with the datapoint ids as keys and
                                                                           //!< the DatapointConfig struct as values

/// Strategy type on how to handle reading the serveritems
enum class ServerItemReadStrategy : std::uint8_t
{
	METADATAREAD,  //!< Read the information for DeviceGroupIdentifier first an then rely on the metadata provided in
	               //!< DevicesMeta.h to read the remaining serveritems
	METADATAREADWITHFALLBACK,  //!< The same as METADATAREAD but instead of throwing an exception if no meta data is
	                           //!< available a fallback to read ALLATONCE is executed
	ALLATONCE                  //!< Read from first to std::uint16_t Max serveritems
};

/// Configuration options for the cpp device class behaviour
struct DeviceClassConfig
{
	bool cacheEnabled;                      //!< Enable the caching of datapoint values and parameter bytes
	bool blockReadEnabled;                  //!< Read in blocks if the request is to larger to handle in one telegram
	ServerItemReadStrategy siReadStrategy;  //!< How server items should be read
};

}  // namespace wz::baos

#endif  // __BAOSLIB_INTERFACE_DATA_TYPES_H__
