//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_TCPCONNECTION_H__
#define __BAOSLIB_TCPCONNECTION_H__

#include "baos/config.h"
#include "baos/connection/BaosConnection.h"

/*!
    @file
    @brief Specialization for a TCP IP v4 BAOSConnection
*/

namespace wz::baos::ip
{

/**********************************************************
 * BaosTcpConnection
 ***********************************************************/

/*!
    @brief A specialization for connection to the BAOS device via a TCP connection
 */
class BAOSLIB_EXPORT BaosTcpConnection : public BaosConnection
{
public:
	using UPtr              = std::unique_ptr<BaosTcpConnection>;  //!< Alias for a std::unique_ptr to this class
	using TcpTransportFrame = Buffer;                              //!< Alias on how a TcpTransportFrame is represented

	BaosTcpConnection();

	/// @cond TESTING
	/*!
	    @brief Construct a new Baos Tcp Connection object with an explicit Ip4Driver to use

	    @param driver
	    @remark Intended for testing only
	 */
	BaosTcpConnection(std::unique_ptr<Ip4Driver>& driver);
	/// @endcond

	virtual ~BaosTcpConnection();

	/*!
	    @brief Open a connection to physical BAOS device

	    @param ip4Address  The IPv4 Address as string in the form "xxx.xxx.xxx.xxx"
	    @param port The tcp port of the remote device
	 */
	void open(const std::string& ip4Address, std::uint16_t port = 12004);
	void reopen() override;

protected:
	bool isOpenImpl() override
	{
		return tcpTransport_.isOpen();
	};

	/*!
	    @brief Wrap a baos frame into a KnxNet/Ip frame

	    @param frame The BaosFrame to wrap
	    @return The wrapped KnxNet/Ip frame
	 */
	TcpTransportFrame wrapKNXnetIp(BaosFrame&& frame);

	protocol::ResponseFuture closeImplAsync() override;
	void startImpl() override;

	/*!
	    @brief Handles the incoming IOEvents
	*/
	void eventHandler();

	/*!
	    @brief Called for every Rx IOEvent (incoming, received)

	    Tries to parse the incoming packet
	    @param event The RX IOEvent with the incming packet as payload
	*/
	void handleRxTcp(IOEvent&& event);

	/*!
	    @brief Called for every Tx IOEvent (outgoing, send, transmit)
	    @param event The TX IOEvent with the packet to send as payload
	 */
	void handleTxTcp(IOEvent&& event);

	protocol::ResponseFuture sendImpl(BaosFrame&& frame) override;
	void onStateChange(State oldState, State newState) override;

private:
	TcpIO tcpTransport_;
	ip::Ip4SocketAddress socketAddress_;
	WZLOGGER("BaosTcpConnection", LVL_DEBUG);
	std::uint64_t msgId_{0};
};

}  // namespace wz::baos::ip

#endif  // __BAOSLIB_TCPCONNECTION_H__
