//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_TCP_IO_H__
#define __BAOSLIB_TCP_IO_H__

#include "baos/BaosLib_export.h"
#include "baos/config.h"
#include "baos/connection/IOEvent.h"

#include "wzcpp/features/concurrent/executor.h"

/// @cond DOCUMENT_STL
#include <array>
#include <atomic>
#include <memory>
#include <string>
/// @endcond


/*!
    @file
    @brief Specialization for a TCP IP v4 BAOSConnection
*/

namespace wz::baos::ip
{

class Ip4Driver;

/*!
    @brief To find packets in the TCP byte stream a pattern and the following toatl packet length are used.

    Provides information about a specific pattern to find packets in the TCP stream
 */
class BAOSLIB_EXPORT PacketizerBase
{
public:
	using UPtr       = std::unique_ptr<PacketizerBase>;  //!< Alias for a std::unique_ptr to this class
	PacketizerBase() = delete;

	/*!
	    @brief Construct a new Packetizer Base object from a given pattern

	    @param startPattern A Buffer containing the bytes which mark the beginning of a packet
	    @param sizePosition  The 0 based index of the total size information in packet
	    @param sizeBytes From how many bytes is the total size composed
	 */
	PacketizerBase(const Buffer& startPattern, std::uint16_t sizePosition, std::uint16_t sizeBytes)
	    : startPattern_{startPattern},
	      sizePosition_{sizePosition},
	      sizeBytes_{sizeBytes}
	{
	}
	virtual ~PacketizerBase() = default;

	/*!
	    @param buf reference to the buffer in which the packet should be found
	    @param bufferSize the actual size of the buffer
	    @param offset [out] if a packet was found the start offset of it
	    @param length [out] if a packet was found the length of the packet(overall)
	    @return true if packet was found false otherwise
	*/
	virtual bool packetize(const Buffer& buf,
	                       std::uint16_t bufferSize,
	                       std::int32_t& offset,
	                       std::uint16_t& length) = 0;

protected:
	const Buffer startPattern_;            //!< The pattern which marks the beginning of a packet in the byte stream
	const std::int32_t sizePosition_{-1};  //!< The 0 based index of the total size information in packet
	const std::uint16_t sizeBytes_{0};     //!< From how many bytes is the total size composed
};

/*!
    @brief Specializaton for a Packetizer which detects KnxNet/Ip packets.
 */
class BAOSLIB_EXPORT KnxNetIPPacketizer : public PacketizerBase
{
public:
	using UPtr = std::unique_ptr<KnxNetIPPacketizer>;  //!< Alias for a std::unique_ptr to this class

	/*!
	@brief Construct a new Knx Net IP Packetizer object with given pattern. size position and length

	 */
	KnxNetIPPacketizer() : PacketizerBase({0x06, 0x20, 0xF0, 0x80}, 4, 2)
	{
	}

	bool packetize(const Buffer& buf, std::uint16_t bufferSize, std::int32_t& offset, std::uint16_t& length) override;
};


/******************************************************************
 * TcpIO
 ******************************************************************/

/*!
    @brief Base class for a TCP input output connection.

    This establishes  a TCP connection to the BAOS devices. It sends and receives raw bytes.
 */
class BAOSLIB_EXPORT TcpIO
{
public:
	using UPtr       = std::unique_ptr<TcpIO>;        //!< Alias for a std::unique_ptr to this class
	using UniqueLock = std::unique_lock<std::mutex>;  //!< Alias for unique lock used by this class

	/*!
	    @brief Construct a new TcpIO object with a given message queue and executor

	    @param msgQueue An external message queue to post events into
	    @param executor An external executor to run this class worker in
	 */
	TcpIO(BaosMsgQueue& msgQueue, wzcpp::ExecutorBase& executor);

	/// @cond TESTING
	/*!
	    @brief Construct a new TcpIO object

	    @param msgQueue An external message queue to post events into
	    @param driver	A given driver to use
	    @param executor An external executor to run this class worker in
	 */
	TcpIO(BaosMsgQueue& msgQueue, std::unique_ptr<Ip4Driver>& driver, wzcpp::ExecutorBase& executor);

	/*!
	    @brief Get the underlying driver

	    @return a reference to the underlying driver
	 */
	Ip4Driver& getDriver()
	{
		return *driver_;
	}
	/// @endcond

	~TcpIO();

	/*!
	    @brief Send a byte list out

	    @param buf The bytes to send
	 */
	void send(const Buffer& buf);

	/*!
	    @brief Open the TCP connection to the remote device

	    @param ip4Address  The IPv4 Address as string in the form "xxx.xxx.xxx.xxx"
	    @param port The tcp port of the remote device
	 */
	void open(const std::string& ip4Address, std::uint16_t port);

	/*!
	    @brief Close the TCP connection
	 */
	void close();

	/*!
	    @brief Wait till the worker of this class has finished
	 */
	void waitForFinished();

	/*!
	    @brief Get the open state of the TCP connection

	    @return true if the TCP connection is established false otherwise
	 */
	bool isOpen()
	{
		return isOpen_;
	};


private:
	void rxHandler();

protected:
	static std::atomic<std::uint64_t> msgId;  //!< Used to generate a continuous unique message id
	std::unique_ptr<Ip4Driver> driver_;       //!< The driver to use for actual sending and receiving
	std::atomic<bool> rxShouldStop_{false};   //!< Flag to indicate that the receiver worker should stop
	std::condition_variable cvFinished_;      //!< Condition variable which is set when the receiver worker has stopped
	mutable std::mutex cvFinished_mut_;       //!< The correponding mutex for the cvFinished_ condition variable
	std::atomic<bool> stopped_{false};  //!< Flag to show the actual state of the receive worker. Because condition
	                                    //!< variables may suffer from spurious wakeups.

	std::atomic<bool> isOpen_{false};  //!< Flag to show if the TCP connection is open or not
	PacketizerBase::UPtr packetizer_;  //!< TCP stream packetizer used by this class
	BaosMsgQueue& msgQueue_;           //!< Reference to the message queue in which events will be posted by this class
	wzcpp::ExecutorBase& executor_;    //!< Reference to the executor to use for running the receive(RX) worker
};

}  // namespace wz::baos::ip

#endif  //__BAOSLIB_TCP_IO_H__
