//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_DEFINES_H__
#define __BAOSLIB_DEFINES_H__

#include "baos/config.h"

/// @cond DOCUMENT_STL
#include <cassert>
#include <map>
#include <sstream>
#include <string>
/// @endcond

/*!
    @file
    @brief Global BAOS protocol defines and types
*/

/// Groups BAOS binary protocol specific types , defines and classes for Indications, Responses etc.
namespace wz::baos::protocol
{
using DatapointValue = wz::baos::Buffer;  //!< An type for the DatapointValues
using DatapointValues =
    std::map<std::uint16_t, DatapointValue>;  //!< A map with datapoint ids as keys and DatapointValue as values

using ServerItem_StatusFlagType = std::uint16_t;  //!< Serveritem status flag type

///  Enum class representing the flags inside the status flag serveritem.
enum class ServerItem_StatusFlag : ServerItem_StatusFlagType
{
	RUNNING      = 1,  //!< The run state of the application
	LOADED       = 2,  //!< The load state of the applcation
	TL_CONNECTED = 4,  //!< If the transport layer is connected or not
	SECURE       = 8   //!< If the device has secure mode enabled
};

/// Requested datapoint value can be filtered by this criteria
enum class DatapointValueFilter : std::uint8_t
{
	ALL,     //!< No filter, all datapoints
	VALID,   //!< Only the valid datapoints. A datapoint is considered valid if the BAOS device has received a value for
	         //!< it
	UPDATED  //!< Retrieve the datapoints that have changed since the last request. If indication sending is enabled NO
	         //!< datapoint get this state, since the indication clears the updated state
};

/// Setting a datapoint value there are different modes supported. This enum class gives those modes names
enum class DatapointSetCommand : std::uint8_t
{
	NoCommand               = 0x00,  //!< Do nothing, more or less a placeholder
	SetNewValue             = 0x01,  //!< Set the new value inside the baos device object server*/
	SendValueOnBus          = 0x02,  //!< Send the new value to bus but does not set it internaly
	SetNewValueAndSendOnBus = 0x03,  //!< The default: Set the new value in the object server and send it to the bus
	ReadNewValueViaBus =
	    0x04,  //!< Issue an group value read on the bus to get the current value of the underlying group object
	ClearTransmissionState = 0x05  //!< Clear the device internal transmission state of the datapoints
};

/// The knx transmit priority
enum class TransmitPriority : std::uint8_t
{
	System = 0x00,
	High   = 0x01,
	Alarm  = 0x02,
	Low    = 0x03
};

BAOSLIB_EXPORT std::string transmitPriorityToString(
    TransmitPriority transmitPrio);  //!< Convert the enum value to a human readable string

///  Internal transmit state of the device
enum class TransmissionStatus : std::uint8_t
{
	Idle_OK    = 0x00,
	Idle_Error = 0x01,
	InProgress = 0x02,
	Request    = 0x03
};

BAOSLIB_EXPORT std::string transmissionStatusToString(
    TransmissionStatus transmissionStatus);  //!< Convert the enum value to a human readable string

/// The in the baos binary protocol specified sub services and some internal ranges in them
enum class BaosSubServiceTypes : std::uint8_t
{
	MinRequest                     = 0x01,
	GetServerItemRequest           = 0x01,
	SetServerItemRequest           = 0x02,
	GetDatapointDescriptionRequest = 0x03,
	GetDescriptionStringRequest    = 0x04,
	GetDatapointValueRequest       = 0x05,
	SetDatapointValueRequest       = 0x06,
	GetParameterByteRequest        = 0x07,
	MaxRequest                     = 0x0C,

	MinResponse                     = 0x81,
	GetServerItemResponse           = 0x81,
	SetServerItemResponse           = 0x82,
	GetDatapointDescriptionResponse = 0x83,
	GetDescriptionStringResponse    = 0x84,
	GetDatapointValueResponse       = 0x85,
	SetDatapointValueResponse       = 0x86,
	GetParameterByteResponse        = 0x87,
	MaxResponse                     = 0x8C,

	MinIndication            = 0xC1,
	DatapointValueIndication = 0xC1,
	ServerItemIndication     = 0xC2,
	MaxIndication            = 0xC2,
};

BAOSLIB_EXPORT std::string getBaosSubServiceTypeDescription(
    BaosSubServiceTypes subServiceType);  //!< Convert the enum value to a human readable string

/// These are the BaosServerErrorCodes
enum class BaosResponseCode : std::uint8_t
{
	Success             = 0,   //!< No error
	Internal            = 1,   //!< Internal error
	NoItemFound         = 2,   //!< No item found
	BufferTooSmall      = 3,   //!< Buffer is too small
	ItemNotWriteable    = 4,   //!< Item is not writable
	ServiceNotSupported = 5,   //!< Service is not supported
	BadServiceParameter = 6,   //!< Bad service parameter
	WrongDpId           = 7,   //!< Wrong datapoint ID
	BadDpCmd            = 8,   //!< Bad datapoint command
	BadDpValueLength    = 9,   //!< Bad length of datapoint value
	MessageInconsistent = 10,  //!< Message inconsistent
	ObjectServerIsBusy  = 11,  //!< Object server is busy
};

std::string baosResponseCodeToString(BaosResponseCode code);  //!< Convert the enum value to a human readable string

/// Enum class representing the different datapoint lengths
enum class DatapointSize : std::uint8_t
{
	BITS_1   = 0,
	BITS_2   = 1,
	BITS_3   = 2,
	BITS_4   = 3,
	BITS_5   = 4,
	BITS_6   = 5,
	BITS_7   = 6,
	BYTE_1   = 7,
	BYTES_2  = 8,
	BYTES_3  = 9,
	BYTES_4  = 10,
	BYTES_6  = 11,
	BYTES_8  = 12,
	BYTES_10 = 13,
	BYTES_14 = 14
};

BAOSLIB_EXPORT std::string dpSizeToString(DatapointSize dpSize);  //!< Convert the enum value to a human readable string

BAOSLIB_EXPORT std::uint8_t dpSizeInBytes(
    DatapointSize dpSize);  //!< Convert the enum value to a intergral size in bytes

//! Enumeration value for the supported datapoint types( DPTs )
enum class DPT : std::uint8_t
{
	DPT_1  = 1,   //!< (1 Bit, Boolean)
	DPT_2  = 2,   //!< (2 Bit, Control)
	DPT_3  = 3,   //!< (4 Bit, Dimming, Blinds)
	DPT_4  = 4,   //!< (8 Bit, Character Set)
	DPT_5  = 5,   //!< (8 Bit, Unsigned Value)
	DPT_6  = 6,   //!< (8 Bit, Signed Value)
	DPT_7  = 7,   //!< (2 std::uint8_t, Unsigned Value)
	DPT_8  = 8,   //!< (2 std::uint8_t, Signed Value)
	DPT_9  = 9,   //!< (2 std::uint8_t, Float Value)
	DPT_10 = 10,  //!< (3 std::uint8_t, Time)
	DPT_11 = 11,  //!< (3 std::uint8_t, Date)
	DPT_12 = 12,  //!< (4 std::uint8_t, Unsigned Value)
	DPT_13 = 13,  //!< (4 std::uint8_t, Signed Value)
	DPT_14 = 14,  //!< (4 std::uint8_t, Float Value)
	DPT_15 = 15,  //!< (4 std::uint8_t, Access)
	DPT_16 = 16,  //!< (14 std::uint8_t, String)
	DPT_17 = 17,  //!< (1 std::uint8_t, Scene Number)
	DPT_18 = 18,  //!< (1 std::uint8_t, Scene Control)
};

BAOSLIB_EXPORT std::string dptToString(DPT dpt);  //!< Convert the enum value to a human readable string


/// An enum class to provide conveniant names to the serveritem ids.
enum class ServerItemId : std::uint8_t
{
	Invalid                   = 0,   //!< Mark item as invalid
	HardwareType              = 1,   //!< (Read Only; 6 bytes)
	HardwareVersion           = 2,   //!< (Read Only; 1 byte)
	FirmwareVersion           = 3,   //!< (Read Only; 1 byte)
	ManufactureCodeDevice     = 4,   //!< (Read Only; 2 bytes)
	ManufactureCodeApp        = 5,   //!< (Read Only; 2 bytes)
	ApplicationId             = 6,   //!< (Read Only; 2 bytes)
	ApplicationVersion        = 7,   //!< (Read Only; 1 bytes)
	SerialNumber              = 8,   //!< (Read Only; 6 bytes)
	TimeSinceReset            = 9,   //!< (Read Only; 4 bytes)
	BusConnected              = 10,  //!< (Read Only; 1 byte)
	MaxBufferSize             = 11,  //!< (Read/Write; 2 bytes)
	LengthOfDescriptionString = 12,  //!< (Read Only; 2 bytes)
	BaudrateItem              = 13,  //!< (Read/Write; 1 byte)
	CurrentBufferSize         = 14,  //!< (Read Only; 2 bytes)
	ProgrammingMode           = 15,  //!< (Read/Write; 1 bytes)
	BinaryProtocolVersion     = 16,  //!< Version of the ObjectServer binary protocol (Read Only; 1 byte; from V2.0)
	IndicationSetting         = 17,  //!< Indicates whether indications are sent (Read/Write; 1 byte; from V2.0)
	WebServicesProtocolVersion =
	    18,  //!< Version of the ObjectServer web services protocol (Read Only; 1 byte; from V2.0)
	RestServiceProtocolVersion = 19,  //!< (Read Only; 1 byte; from V2.1)
	IndividualAddress          = 20,  //!< (Read / Write; 2 byte; from V2.1)
	MacAddress                 = 21,  //!< (Read Only; 6 byte; from V2.1)
	TunnellingEnabled          = 22,  //!< (Read / Write; 1 byte; from V2.1)
	BaosBinaryEnabled          = 23,  //!< (Read / Write; 1 byte; from V2.1)
	BaosWebEnabled             = 24,  //!< (Read / Write; 1 byte; from V2.1)
	BaosRestEnabled            = 25,  //!< (Read / Write; 1 byte; from V2.1)
	HttpFileEnabled            = 26,  //!< (Read / Write; 1 byte; from V2.1)
	SearchRequestEnabled       = 27,  //!< (Read / Write; 1 byte; from V2.1)
	IsStructured               = 28,  //!< (Read Only; 1 byte; from V2.1)
	MaxManagementClients       = 29,  //!< (Read Only; 1 byte; from V2.1)
	ConnectedManagementClients = 30,  //!< (Read Only; 1 byte; from V2.1)
	MaxTunnellingClients       = 31,  //!< (Read Only; 1 byte; from V2.1)
	ConnectedTunnellingClients = 32,  //!< (Read Only; 1 byte; from V2.1)
	MaxBaosUdpClients          = 33,  //!< (Read Only; 1 byte; from V2.1)
	ConnectedBaosUdpClients    = 34,  //!< (Read Only; 1 byte; from V2.1)
	MaxBaosTcpClients          = 35,  //!< (Read Only; 1 byte; from V2.1)
	ConnectedBaosTcpClients    = 36,  //!< (Read Read / Write Only; 1 byte; from V2.1)
	DeviceFriendlyName         = 37,  //!< (Read / Write; 30 bytes; from V2.1)
	MaxDatapoints              = 38,  //!< (Read Only; 2 bytes; from V2.1)
	ConfiguredDatapoints       = 39,  //!< (Read Only; 2 bytes; from V2.1)
	MaxParameterBytes          = 40,  //!< (Read Only; 2 bytes; from V2.1)
	DownloadCounter            = 41,  //!< (Read Only; 2 bytes; from V2.1)
	IPAssignment               = 42,  //!< (Read / Write; 1 byte; from V2.1)
	IPAddress                  = 43,  //!< (Read / Write; 4 bytes; from V2.1)
	SubnetMask                 = 44,  //!< (Read / Write; 4 bytes; from V2.1)
	DefaultGateway             = 45,  //!< (Read / Write; 4 bytes; from V2.1)
	TimeSinceResetUnit         = 46,  //!< (Read / Write; 1 byte; from V2.1)
	SystemTime                 = 47,  //!< (Read / Write; variable length; from V2.1)
	SystemTimezoneOffset       = 48,  //!< (Read / Write; 1 byte (signed); from V2.1)
	MenuEnabled                = 49,  //!< (Read / Write; 1 byte; from V2.1)
	SuspendEnabled             = 50,  //!< (Read / Write; 1 byte; from 2.1)
	RFDomainAddress            = 51,  //!< (Read / Write; 6 bytes;)
	StatusFlagsSupported       = 52,  //!< (Read Only; 2 bytes; from V2.1)
	StatusFlags                = 53   //!< (Read Only; 2 bytes; from V2.1)
};

BAOSLIB_EXPORT std::string serverItemIdToString(
    ServerItemId sid);  //!< Convert the enum value to a human readable string

/// This structure represents the status flags provided by some BAOS devices server item ServerItemId::StatusFlags
class BAOSLIB_EXPORT StatusFlags
{
public:
	/// Construct from a existing ServerItem_StatusFlagType value
	StatusFlags(const ServerItem_StatusFlagType val) : value_{val} {};
	/// Construct from an array of bytes
	StatusFlags(const Buffer& data)
	{
		assert(data.size() >= 2);
		value_ = data[0] << 8 | data[1];
	}

	/*!
	    @brief Construct a new Status Flags object

	    @param running The run state of the application
	    @param loaded  The load sate of the application
	    @param tl_connected Connection state of the transport layer
	    @param secure  Data security enabled
	 */
	StatusFlags(bool running, bool loaded, bool tl_connected, bool secure)
	{
		value_ = 0x00;
		if (running)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::RUNNING);
		if (loaded)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::LOADED);
		if (tl_connected)
			value_ |=
			    static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::TL_CONNECTED);
		if (secure)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::SECURE);
	}

	/// Return the value of the running element as boolean
	bool isRunning() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::RUNNING);
	}

	/// Return if the ETS applicatiopn is loaded or not
	bool isLoaded() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::LOADED);
	}

	/// Returns if the transport layer is connected
	bool isTLConnected() const
	{
		return value_ &
		       static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::TL_CONNECTED);
	}

	/// Returns if the device is on secure mode
	bool isSecure() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::SECURE);
	}

	/// Conversion operator to ServerItem_StatusFlagType
	operator ServerItem_StatusFlagType() const
	{
		return value_;
	}

private:
	ServerItem_StatusFlagType value_;  //!< The internal value storage
};


/// Represents which status flags are supported by the device, serveritem 52
class BAOSLIB_EXPORT StatusFlagsSupported
{
public:
	//!< Construct from a existing ServerItem_StatusFlagType value
	StatusFlagsSupported(const ServerItem_StatusFlagType val) : value_{val} {};

	/// Construct from an array of bytes
	StatusFlagsSupported(const Buffer& data)
	{
		assert(data.size() >= 2);
		value_ = data[0] << 8 | data[1];
	}

	/*!
	    @brief Construct a new StatusFlagsSupported object from integral values

	    @param running The run state of the application
	    @param loaded  The load sate of the application
	    @param tl_connected Connection state of the transport layer
	    @param secure  Data security enabled
	 */
	StatusFlagsSupported(bool running, bool loaded, bool tl_connected, bool secure)
	{
		value_ = 0x00;
		if (running)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::RUNNING);
		if (loaded)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::LOADED);
		if (tl_connected)
			value_ |=
			    static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::TL_CONNECTED);
		if (secure)
			value_ |= static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::SECURE);
	}

	/// Is the running flag supported
	bool hasRunning() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::RUNNING);
	}

	/// Is the loaded flag supported
	bool hasLoaded() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::LOADED);
	}

	/// Is the transport layer connected flag supported
	bool hasTLConnected() const
	{
		return value_ &
		       static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::TL_CONNECTED);
	}

	/// Is the secure mode flag supported
	bool hasSecure() const
	{
		return value_ & static_cast<std::underlying_type<ServerItem_StatusFlag>::type>(ServerItem_StatusFlag::SECURE);
	}

	/// Conversion operator to ServerItem_StatusFlagType type
	operator ServerItem_StatusFlagType() const
	{
		return value_;
	}

private:
	ServerItem_StatusFlagType value_;  //!< The internal value storage
};

/// Structure to hold information about a serveritem
struct BAOSLIB_EXPORT ServerItem
{
	ServerItem() = default;  // The default constructor

	/*!
	    @brief Construct a new Server Item object

	    @param id_  the serveritem id
	    @param length_  the length of the data
	    @param data_  a buffer containing the data bytes
	 */
	ServerItem(ServerItemId id_, std::uint8_t length_, Buffer&& data_) : id{id_}, length{length_}, data{data_}
	{
	}
	ServerItemId id{ServerItemId::Invalid};  //!< the serveritem id
	std::uint8_t length{0};                  //!< the length of the data
	Buffer data;                             //!< the data
};

using ServerItems = std::vector<ServerItem>;  //!< Type for a list of ServerItems

/// Struct for the communication flags of a datapoint. See KNX spec for more details
struct BAOSLIB_EXPORT DatapointConfigFlags
{
	DatapointConfigFlags()
	    : transmitPriority{0},
	      communication{0},
	      readFromBus{0},
	      writeFromBus{0},
	      readOnInit{0},
	      transmitToBus{0},
	      updateResponse{0}
	{
	}

	/// Construct from a byte value
	DatapointConfigFlags(std::uint8_t&& value)
	{
		updateResponse   = value >> 7;
		transmitToBus    = value >> 6 & 0x01;
		readOnInit       = value >> 5 & 0x01;
		writeFromBus     = value >> 4 & 0x01;
		readFromBus      = value >> 3 & 0x01;
		communication    = value >> 2 & 0x01;
		transmitPriority = value & 0x03;
	}

	/// Assign a byte to this object
	DatapointConfigFlags& operator=(std::uint8_t& value)
	{
		updateResponse   = value >> 7;
		transmitToBus    = value >> 6 & 0x01;
		readOnInit       = value >> 5 & 0x01;
		writeFromBus     = value >> 4 & 0x01;
		readFromBus      = value >> 3 & 0x01;
		communication    = value >> 2 & 0x01;
		transmitPriority = value & 0x03;
		return *this;
	}

	/// Conversion operator to std::uint8_t
	operator std::uint8_t() const
	{
		std::uint8_t value{0};
		value = updateResponse << 7 | transmitToBus << 6 | readOnInit << 5 | writeFromBus << 4 | readFromBus << 3 |
		        communication << 2 | transmitPriority;
		return value;
	}

	/// Convert to a string representation of the communication flags, e.g. High CRWTUI every flag not set is
	/// represented by an underscore
	std::string toString() const
	{
		std::ostringstream oss;
		std::string flags{"______"};
		if (communication) flags[0] = 'C';
		if (readFromBus) flags[1] = 'R';
		if (writeFromBus) flags[2] = 'W';
		if (transmitToBus) flags[3] = 'T';
		if (updateResponse) flags[4] = 'U';
		if (readOnInit) flags[5] = 'I';

		oss << transmitPriorityToString(static_cast<TransmitPriority>(transmitPriority)) << " " << flags;
		return oss.str();
	}

	std::uint8_t transmitPriority : 2,  //!< The KNX transmit priority
	    communication : 1,   //!< The communication flag, without it no KNX Bus communication is possible for the group
	                         //!< object, will be cleared by the stack of no group address is assigned.
	    readFromBus : 1,     //!< The readable from bus flag
	    writeFromBus : 1,    //!< The writable from bus flag
	    readOnInit : 1,      //!< The perform read request in onit flag
	    transmitToBus : 1,   //!< The devices send changes on the group object to the KNX bus
	    updateResponse : 1;  //!< Define the bitfield for the flags and transmitPriority
};

/// Internal state for each datapoint in the object server
struct BAOSLIB_EXPORT DatapointStatus
{
	DatapointStatus()
	    : transmitStatus{0},
	      readRequest{0},
	      updated{0},
	      valid{0},
	      reserved_1{0},
	      reserved_2{0},
	      reserved_3{0}
	{
	}

	/// Construct from byte value
	DatapointStatus(std::uint8_t&& value)
	{
		reserved_3     = value >> 7;
		reserved_2     = value >> 6 & 0x01;
		reserved_1     = value >> 5 & 0x01;
		valid          = value >> 4 & 0x01;
		updated        = value >> 3 & 0x01;
		readRequest    = value >> 2 & 0x01;
		transmitStatus = value & 0x03;
	}

	/// Assign value from byte
	DatapointStatus& operator=(std::uint8_t& value)
	{
		reserved_3     = value >> 7;
		reserved_2     = value >> 6 & 0x01;
		reserved_1     = value >> 5 & 0x01;
		valid          = value >> 4 & 0x01;
		updated        = value >> 3 & 0x01;
		readRequest    = value >> 2 & 0x01;
		transmitStatus = value & 0x03;
		return *this;
	}

	/// Conversion operator to std::uint8_t
	operator std::uint8_t() const
	{
		std::uint8_t value{0};
		value = reserved_3 << 7 | reserved_2 << 6 | reserved_1 << 5 | valid << 4 | updated << 3 | readRequest << 2 |
		        transmitStatus;
		return value;
	}

	/*!
	    @brief  Represent the flags in a human readable form

	    The output string has the form e.g. "VUR  Idle_OK" every flag not set is represented by an underscore.
	    The first character is V for valid, the second is U for updated, the third is R for read request and W for write
	   request. The transmission status is represented by @see transmissionStatusToString
	    @return std::string
	 */
	std::string toString() const
	{
		std::ostringstream oss;
		valid ? oss << "V" : oss << "_";
		updated ? oss << "U" : oss << "_";
		readRequest ? oss << "R" : oss << "W";
		oss << " ";
		oss << transmissionStatusToString(static_cast<TransmissionStatus>(transmitStatus));
		return oss.str();
	}

	//@{
	/*!
	    @brief See BAOS binary protocol spec GetDatapointValue.Res state byte for details
	 */
	std::uint8_t transmitStatus : 2,  //!< The tranmission status. @see TransmissionStatus
	    readRequest : 1,              //!< 0 Write request should be sent, 1 Read request should be sent
	    updated : 1,  //!< Value was updated from bus, cleared after it is read be client or indication is send to
	                  //!< client
	    valid : 1,    //!< 0 Object value is invalid or 1 the device has seen at least one value for the group object.
	    reserved_1 : 1,  //!< Reserved bit 1
	    reserved_2 : 1,  //!< Reserved bit 2
	    reserved_3 : 1;  //!< Reserved bit 3
	                     //@}
};

/// A combination of the status of a datapoint and its actual value
struct BAOSLIB_EXPORT DatapointValueState
{
	DatapointValue value;    //!< The actual value of the datapoint
	DatapointStatus status;  //!< The status of the datapoint object
};

using DatapointValueStates = std::map<std::uint16_t, DatapointValueState>;  //!< A map between the datapoint id as key
                                                                            //!< and the DatapointValueState as value

/// A pair of the DatapointSetCommand and the Datapoint value for changing datapoint values.
struct BAOSLIB_EXPORT DatapointChangeValue
{
	DatapointValue value;         //!< The actual value of the datapoint
	DatapointSetCommand command;  //!< The command for the value change
};

using DatapointChangeValues = std::map<std::uint16_t, DatapointChangeValue>;  //!< A map between the datapoint id as key
                                                                              //!< and the DatapointChangeValue as value

/// Represents the information about a datapoint in the BAOS device
struct BAOSLIB_EXPORT DatapointDescription
{
	DatapointSize size;          //!< The size as an enum value
	DatapointConfigFlags flags;  //!< The communication flags
	DPT dpt;  //!< The datapoint type (e.G. DPT1 for boolean, see KNX Spec datapoint types for more details)
};

using DatapointDescriptions =
    std::map<std::uint16_t, DatapointDescription>;  //!< A map between the datapoint id as key and the
                                                    //!< DatapointDescription structs as value
using DatapointDescriptionString = std::string;     //!< Strong type for the description string
using DatapointDescriptionStrings =
    std::map<std::uint16_t, DatapointDescriptionString>;  //!<  A map between the datapoint id as key and the
                                                          //!<  DatapointDescriptionStrings as value

using ParameterBytes = std::vector<std::uint8_t>;  //!< Strong type for a list of parameter bytes

/// The ServerItemId::TimeSinceResetUnit current unit for TimeSinceReset. This type gives those units a name.
enum class TimeSinceResetUnit
{
	milliseconds,
	seconds,
	minutes,
	hours
};

BAOSLIB_EXPORT std::string timeSinceResetUnitToString(
    TimeSinceResetUnit unit);  //!< Convert the enum value to a human readable string

/// An Ip device may have different methods for ip assigment
enum class IpAssigmentMethod
{
	manual = 0x1,  //!< The ip address, subnet mask etc. are explicitly set manualy
	bootp  = 0x2,  //!< Boot mode with network boot vi bootp
	dhcp   = 0x4,  //!< The ip address is requested from a DHCP server
	autoIp = 0x8   //!< Configure its own ip address automatically
};


/*!
    @name Protocol defines and positions
*/
///@{
constexpr std::uint8_t BAOS_MAIN_SERVICE = 0xF0;  //!< The main service to identify a BAOS telegram by
constexpr std::size_t POS_MAIN_SERVICE   = 0;     //!< The 0 based position of the main service byte
constexpr std::size_t POS_SUB_SERVICE    = 1;     //!< The 0 based position of the sub service byte
constexpr std::size_t POS_START =
    2;  //!< The 0 based position of the start element (startitem, start datapoint id etc.)
constexpr std::size_t POS_NUMBER_OF_ELEMENTS = 4;  //!< The 0 based position of the number of elements in the packet
constexpr std::size_t POS_ERRORCODE          = 6;  //!< The 0 bases position of the errorcode if available
constexpr std::size_t POS_PAYLOAD_BEGIN      = 6;  //!< The 0 based position of the payload in the packet
///@}

}  // namespace wz::baos::protocol

#endif
