//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_INDICATION_H__
#define __BAOSLIB_INDICATION_H__

#include "baos/BaosLib_export.h"
#include "baos/config.h"
#include "baos/protocol/Defines.h"
#include "baos/protocol/Response.h"
#include "wzcpp/features/concurrent/concurrent_queue.h"

/// @cond DOCUMENT_STL
#include <future>
#include <memory>
#include <variant>
/// @endcond


/*!
    @file
    @brief Global BAOS indication classes and defines
*/

namespace wz::baos::protocol
{

using IndPromise = std::promise<protocol::BaseResponse>;  //!< Strong type for a indication promise
using IndFuture  = std::future<protocol::BaseResponse>;   //!< Strong type for a indication future


/*!
    @brief An Indication event

    Represents in indication with its type and payload
 */
class BAOSLIB_EXPORT IndicationEvent
{
public:
	using UPtr = std::unique_ptr<IndicationEvent>;  //!< Alias for an unique pointer from this class
	using EventDataType =
	    std::variant<protocol::ServerItems, protocol::DatapointValueStates>;  //!< The possible different payload types

	/// The indication types
	enum class Type
	{
		SERVERITEM,  //!< A server item indication
		DP_VALUE,    //!< A datapoint value change indication
		UNKNOWN      //!< An indication of unknown type
	};
	IndicationEvent() = default;  //!< The default constructor

	/*!
	    @brief Construct a new Indication Event object

	    @param eventT The IndicationEvent::Type
	    @param pkt The paylod / data
	    @param prom The promise related to this indication
	    @param mID An unqiue message id for this indication
	 */
	IndicationEvent(Type eventT, EventDataType&& pkt, IndPromise&& prom, std::uint64_t mID = 0)
	    : eventType{eventT},
	      msgID{mID},
	      indPromise{std::move(prom)}
	{
		data_ = std::move(pkt);
	}

public:
	Type eventType{Type::UNKNOWN};  //!< The indication type
	std::uint64_t msgID{0};         //!< The unqiue message id
	IndPromise indPromise;          //!< The indication promise

	// private:
	EventDataType data_;  //!< The payload
};


/// Convert IndicationEvent into human readable string
inline std::string toString(IndicationEvent::Type typ)
{
	switch (typ)
	{
		case IndicationEvent::Type::SERVERITEM:
			return "SERVERITEM";
		case IndicationEvent::Type::DP_VALUE:
			return "DP_VALUE";
		default:
			return "Unhandled IOEvent type";
	}
}


using BaosIndQueue =
    wzcpp::ConcurrentDeque<IndicationEvent>;  //!< Alias for a wzcpp::ConcurrentDeque holding indication events

}  // namespace wz::baos::protocol

#endif  //__BAOSLIB_INDICATION_H__
