//
// Copyright (c) 2021 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_BAOSRESPONSE_H__
#define __BAOSLIB_BAOSRESPONSE_H__

#include "baos/Exception.h"
#include "baos/protocol/Defines.h"

/// @cond DOCUMENT_STL
#include <future>
#include <memory>
#include <sstream>
/// @endcond

/*!
    @file
    @brief Classes regarding response telegrams
 */

namespace wz::baos::protocol
{

/// Baseclass for generic responses.
struct BaseResponse
{
	using SPtr = std::shared_ptr<BaseResponse>;  //!< An alias for a std::shared pointer holding an BaseResponse class
	BaseResponse()          = default;
	virtual ~BaseResponse() = default;
};


///  Baseclass for BAOS Responses. Containes the response code and optional in derived classes the payload
struct BaosBaseResponse : public BaseResponse
{
	using SPtr =
	    std::shared_ptr<BaosBaseResponse>;  //!< An alias for a std::shared pointer holding an BaosBaseResponse class

	BaosBaseResponse() = default;

	/*!
	    @brief Construct a new BaosBaseResponse object

	    @param code_  From a BAOS response code. @see BaosResponseCode
	 */
	BaosBaseResponse(protocol::BaosResponseCode code_) : code{code_} {};
	virtual ~BaosBaseResponse() = default;

	protocol::BaosResponseCode code{protocol::BaosResponseCode::Success};  //!< holds the BaosResponse code
};


///  An actual BAOS response with payload of template parameter T
template<class T>
struct BaosResponse : public BaosBaseResponse
{
	BaosResponse() = default;
	/*!
	    @brief Construct a new BaosResponse object

	    @param code_  The BaosResponse code
	    @param items The payload items
	*/
	BaosResponse(protocol::BaosResponseCode code_, T&& items) : BaosBaseResponse(code_), payload_{items} {};

	virtual ~BaosResponse() = default;
	using SPtr = std::shared_ptr<BaosResponse>;  //!< An alias for a std::shared pointer holding an BaosResponse class

	/// Getter for payload
	T getPayload()
	{
		return payload_;
	}  // getter

	/// Setter for payload
	void setPayload(T&& value)
	{
		payload_ = value;
	}  // setter
private:
	T payload_;
};


using ResponsePromise = std::promise<BaseResponse::SPtr>;  //!< Strong type for a promise containing a shared pointer to
                                                           //!< a BaseResponse or any from it derived class
using ResponseFuture = std::future<BaseResponse::SPtr>;  //!< Strong type for a future containing a shared pointer to a
                                                         //!< BaseResponse or any from it derived class


/*!
    @brief Helper function to wait for a ResponseFuture to become ready and handle it

    @tparam T  The Type of the payload in the response
    @param fut An rvalue reference of the response future to handle
    @return T  Returns the payload of type T
 */
template<class T>
T waitAndHandleBaosResponse(ResponseFuture&& fut)
{
	const auto response = fut.get();  // Wait for the response to become ready
	auto responseBaos   = std::dynamic_pointer_cast<BaosBaseResponse>(
        response);  // get the value from the future and cast it to the base type
	if (responseBaos->code == protocol::BaosResponseCode::NoItemFound)
	{
		return T();  // Return empty result
	}
	else if (responseBaos->code != protocol::BaosResponseCode::Success)
	{
		std::ostringstream oss;
		oss << "Error response from BAOS " << protocol::baosResponseCodeToString(responseBaos->code);
		throw wz::baos::WzResponseException(WZ_SL, oss.str().c_str(), static_cast<std::uint32_t>(responseBaos->code));
	}
	else
	{
		auto responseDerived = std::dynamic_pointer_cast<BaosResponse<T>>(
		    response);  // get the value from the future and cast it to the actual type
		return responseDerived->getPayload();
	}
}

/// Handle responses with no payload
void waitAndHandleEmptyBaosResponse(ResponseFuture&& fut);

}  // namespace wz::baos::protocol

#endif
