//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_BLOCKREADER_H__
#define __BAOSLIB_BLOCKREADER_H__

#include "baos/config.h"

#include "wzcpp/features/algorithm/append.h"
#include "wzcpp/features/logging.h"
#include "wzcpp/features/utils/ByteArray.h"
#include "wzcpp/features/utils/Memory.h"

#include <memory>
#include <optional>
#include <utility>

namespace wz::utility
{

/**********************************************
 * Block reader function
 *********************************************/

/*!
    @brief Calls a function very every block in a lager range

    For a given range of elements beginning with start and a given count, split this range into blocks of
    blockSize size and invoke for every block the callback cb.
    Combines the result of every callback into overall result and return it.
*/
class BlockReader
{
public:
	template<class TCallback>
	static auto read(std::uint16_t start, std::uint16_t count, std::uint16_t blockSize, TCallback cb)
	{
		using ReturnTypeOfCb = typename decltype(cb)::result_type;
		WZLOGGER("BlockReader.read", LVL_DEBUG);
		ReturnTypeOfCb result;

		std::uint16_t overallEnd = start + count - 1;
		wzlogs_d("blockReader reading from " << start << " - " << overallEnd << " in blocks of " << blockSize);

		for (std::uint16_t blockStart = start; blockStart < start + count; blockStart += blockSize)
		{
			auto currentCount = blockSize;
			if (currentCount > overallEnd - blockStart + 1)
			{
				currentCount = overallEnd - blockStart + 1;
				wzlogs_d("Current block size limited to " << currentCount);
			}
			wzlogs_d("Reading block from " << blockStart << " to " << blockStart + currentCount - 1
			                               << " (count:" << currentCount << ")");
			wzcpp::append(cb(blockStart, currentCount), result);
		}
		return result;
	}
};

}  // namespace wz::utility

#endif  // __BAOSLIB_BLOCKREADER_H__
