//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __BAOSLIB_TCP_DRIVER_MOCK_H__
#define __BAOSLIB_TCP_DRIVER_MOCK_H__

#include "TcpDriver.h"
#include "baos/BaosLib_export.h"
#include "baos/config.h"
#include <atomic>
#include <condition_variable>
#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <thread>


/// @cond TESTING
namespace wz::baos::ip
{

/*!
   A simple mock class that gets a list of response telegrams and on every read attempt the next element in the list is
   returned until the list is exhaused.
*/
class BAOSLIB_EXPORT Ip4TcpMockDriver : public Ip4Driver
{
public:
	Ip4TcpMockDriver();
	~Ip4TcpMockDriver();
	void open(const std::string& IP4Address, std::uint16_t port) override;
	void close() override;
	std::uint16_t read(Buffer& buf, std::uint16_t offset = 0) override;
	void write(const Buffer& buf) override;
	bool isOpen() override;

	void setReadTelegrams(std::vector<Buffer>&& telegrams)
	{
		recvTelegrams_   = telegrams;
		currentReadIndex = 0;
	}

public:
	Buffer lastSend;
	Buffer lastRead;
	std::size_t currentReadIndex{0};

private:
	std::thread rxThread_;
	bool isOpen_{false};
	std::vector<Buffer> recvTelegrams_;
};

/*!
   A simple mock class that gets a list of request response pairs and for every request matching one in it the
   correspondung response is returned returned until the list is exhaused.
*/
class BAOSLIB_EXPORT Ip4TcpMockRespDriver : public Ip4Driver
{
public:
	using RequestResponses = std::map<Buffer, Buffer>;  // Pairs of requests and responses
	using UniqueLock       = std::unique_lock<std::mutex>;

	Ip4TcpMockRespDriver();
	~Ip4TcpMockRespDriver();
	void open(const std::string& IP4Address, std::uint16_t port) override;
	void close() override;
	std::uint16_t read(Buffer& buf, std::uint16_t offset = 0) override;
	void write(const Buffer& buf) override;
	bool isOpen() override;

	void setRequestResponses(RequestResponses&& requestResponses)
	{
		requestResponses_ = std::move(requestResponses);
	}

public:
	Buffer lastSend;
	Buffer lastRead;
	Buffer currentResponse;

private:
	std::condition_variable cv_;
	mutable std::mutex cvMutex_;
	std::atomic<bool> shouldStop_;
	bool isOpen_{false};
	mutable std::mutex reqResMutex_;
	RequestResponses requestResponses_;
	std::atomic<bool> responseAvailable_;
};

}  // namespace wz::baos::ip

/// @endcond

#endif  // __BAOSLIB_TCP_DRIVER_MOCK_H__
