//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosIpv4Device.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>

#include <iostream>
#include <memory>

/*!
    @example DatapointValueIndication.cpp

    This sample shows how to register a callback function
    for datapoint value indications via the connection API.

    Usage:
    @code
        DatapointValueIndication <ip_address>
        // e.g.:
        DatapointValueIndication 10.0.0.102
    @endcode
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Anonymous namespace
*******************************************/

namespace
{

/*!
    Trace the parameter bytes
*/
void traceDatapointValue(const DatapointValue& dpValue)
{
	std::cout.fill('0');
	std::cout << std::hex << std::uppercase;
	for (auto& item : dpValue)
	{
		std::cout << std::setw(2) << static_cast<int>(item) << " ";
	}
	std::cout << std::nouppercase << std::dec;
}

}  // end anonymous namespace

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "DatapointValueIndication sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles also the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection with the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);

		// register callback function for datapoint value indications
		connection.setDatapointValueIndCallback(
		    [&connection](const DatapointValueStates& valueStates)
		    {
			    for (const auto& item : valueStates)
			    {
				    const DatapointValueState& value = item.second;

				    std::cout << "Received datapoint value indication for id " << item.first;
				    std::cout << ", value: ";
				    traceDatapointValue(value.value);
				    std::cout << std::endl;
			    }
		    });

		// wait for an Enter, all indications will be shown
		// in the callback handler
		std::cout << "Press [Enter] to exit the application ..." << std::endl;
		std::getchar();
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
