//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>
#include <wzcpp/features/utils.h>

#include <iostream>
#include <memory>

/*!
    @example GetDatapointDescriptionString.cpp

    This sample shows how to read datapoint description strings via the connection API <br>

    Usage:
    @code
        GetDatapointDescriptionString <ip_address>
        // e.g.:
        GetDatapointDescriptionString 10.0.0.102
    @endcode

    @note Another possibility is via the device API. <br>
    See also: Sample 03_BaosDatapointConfiguration.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "GetDatapointDescriptionString sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);

		// Reads some datapoint descriptions from the device (single item per requested)

		// Read datapoint description string with id 1
		std::cout << "Reading description string from dp id 1 ... " << std::endl;
		auto desc = connection.getDpDescriptionString(1);
		if (desc.has_value())
		{
			std::cout << "DP 1: " << *desc << std::endl;
		}
		std::cout << std::endl;

		// Read datapoint description string with id 10
		std::cout << "Reading description string from dp id 10 ... " << std::endl;
		desc = connection.getDpDescriptionString(10);
		if (desc.has_value())
		{
			std::cout << "DP 10: " << *desc << std::endl;
		}
		std::cout << std::endl;

		// Reads multiple datapoint description strings (multiple items per requested)

		// start = 1; number of = 50
		std::cout << "Reading description string from dp id 1 to 50 ... " << std::endl;
		const DatapointDescriptionStrings descriptions = connection.getDpDescriptionStrings(1, 50);

		// trace the descriptions
		for (const auto& item : descriptions)
		{
			std::cout << "Datapoint id " << item.first;

			const DatapointDescriptionString& desc = item.second;
			std::cout << ", Description string: " << desc << std::endl;
		}
			std::cout << std::endl;
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
