//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>
#include <wzcpp/features/utils.h>

#include <iostream>
#include <memory>

/*!
    @example GetParameterByte.cpp

    This sample shows how to read paramter bytes via he connection API <br>

    Usage:
    @code
        GetParamterByte <ip_address>
        // e.g.:
        GetParamterByte 10.0.0.102
    @endcode

    @note Another possibility is via the device API. <br>
    See also: Sample 02_BaosParameterBytes.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Anonymous namespace
*******************************************/

namespace
{

/*!
    Trace the parameter bytes
*/
void traceParameterBytes(const ParameterBytes& parameterBytes)
{
	std::cout.fill('0');
	std::cout << std::hex << std::uppercase;
	for (auto& item : parameterBytes)
	{
		std::cout << std::setw(2) << static_cast<int>(item) << " ";
	}
	std::cout << std::nouppercase << std::dec;
}

}  // end anonymous namespace

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "GetParamterByte sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);


		// Reads some parameter bytes (single item per requested)

		// index = 1
		std::cout << "Reading parameter byte 1:" << std::endl;
		auto parameterByte = connection.getParameterByte(1);
		if (parameterByte.has_value())
		{
			std::cout << "Parameter data (1): " << std::setw(2) << static_cast<int>(*parameterByte) << std::endl;
		}
		std::cout << std::endl;

		// index = 12
		std::cout << "Reading parameter byte 12:" << std::endl;
		parameterByte = connection.getParameterByte(12);
		if (parameterByte.has_value())
		{
			std::cout << "Parameter data (12): " << std::setw(2) << static_cast<int>(*parameterByte) << std::endl;
		}
		std::cout << std::endl;


		// Reads multiple parameter bytes (multiple items per requested)

		// index = 2; number of = 4
		std::cout << "Reading parameter bytes 2 - 5:" << std::endl;
		ParameterBytes parameterBytes = connection.getParameterBytes(2, 4);
		traceParameterBytes(parameterBytes);
		std::cout << std::endl << std::endl;

		// index = 11; number of = 50
		std::cout << "Reading parameter bytes 11 - 60:" << std::endl;
		parameterBytes = connection.getParameterBytes(11, 50);
		traceParameterBytes(parameterBytes);
		std::cout << std::endl << std::endl;
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
