//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosIpv4Device.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>

#include <iostream>
#include <memory>

/*!
    @example ServerItemIndication.cpp

    This sample shows how to register a callback function
    for server item indications via the connection API.

    Usage:
    @code
        ServerItemIndication <ip_address>
        // e.g.:
        ServerItemIndication 10.0.0.102
    @endcode
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "ServerItemIndication sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);

		// register callback function for server item indications
		connection.setServerItemsIndCallback(
		    [&connection](const ServerItems& serverItems)
		    {
			    for (const auto& serverItem : serverItems)
			    {
				    std::cout << "Received server item indication for id " << static_cast<unsigned int>(serverItem.id);
				    const std::string description = serverItemIdToString(serverItem.id);
				    std::cout << ", " << description;
				    std::cout << ", data: ";
				    std::cout.fill('0');
				    std::cout << std::hex << std::uppercase;
				    for (auto& value : serverItem.data)
				    {
					    std::cout << std::setw(2) << static_cast<int>(value) << " ";
				    }
				    std::cout << std::nouppercase << std::dec << std::endl;
			    }
		    });

		// wait for an Enter, all indications will be shown
		// in the callback handler
		std::cout << "Press [Enter] to exit the application ..." << std::endl;
		std::getchar();
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
