//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>
#include <wzcpp/features/utils.h>

#include <iostream>
#include <memory>

/*!
    @example SetDatapointValue.cpp

    This sample shows how to set datapoint values via the connection API <br>


    Datapoint

    Usage:
    @code
        SetDatapointValue <ip_address>
        // e.g.:
        SetDatapointValue 10.0.0.102
    @endcode

    Pre-condition:
    It assumes that the following datapoints are configured:

    @verbatim

    |---------|-----------------|-----------------------------|-------------|---------------|
    | Number  |  DPT            |  Name                       |  SizeInBit  |  Flags        |
    |         |                 |                             |             | C R W T U I   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    1    | DPT 01 - 1 bit  | Actuator A - Switch         | 1           | C - W - - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    2    | DPT 01 - 1 bit  | Actuator A - State          | 1           | C R - T - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    3    | DPT 01 - 1 bit  | Actuator B - Switch         | 1           | C - W - - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    |    4    | DPT 01 - 1 bit  | Actuator B - State          | 1           | C R - T - -   |
    |---------|-----------------|-----------------------------|-------------|---------------|
    @endverbatim

    @note Another possibility is via the device API. <br>
    See also: Sample 04_BaosDatapointValues.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "SetDatapointValue sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// Create a TPC/IP connection to the remote BAOS device
		BaosTcpConnection connection;
		connection.open(ipAddress);


		// Set some datapoint values (single item per request)

		// id = 1 (DPT 01 - 1 bit)
		connection.setDpValue(1, DatapointChangeValue{{0x00}, DatapointSetCommand::SetNewValue});

		// Set multiple datapoint values (multiple items per request)

		// id = 1, 2, 3 and 4
		connection.setDpValues(
		    DatapointChangeValues{{1, DatapointChangeValue{{0x01}, DatapointSetCommand::SetNewValue}},
		                          {2, DatapointChangeValue{{0x01}, DatapointSetCommand::SetNewValueAndSendOnBus}},
		                          {3, DatapointChangeValue{{0x00}, DatapointSetCommand::SetNewValue}},
		                          {4, DatapointChangeValue{{0x00}, DatapointSetCommand::SetNewValueAndSendOnBus}}});
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
