//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"
#include <baos/Sdk.h>
#include <baos/ip/BaosIpv4Device.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>

#include <iostream>
#include <memory>

/*!
    @example 02_BaosParameterBytes.cpp

    This sample shows how to read paramter byte via the device API.

    Usage:
    @code
        02_BaosParameterBytes <ip_address>
        // e.g.:
        02_BaosParameterBytes 10.0.0.102
    @endcode

    @note Another possibility is via the connection API. <br>
    See also: Sample GetParameterByte.cpp
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Anonymous namespace
*******************************************/

namespace
{

/*!
    Trace the parameter bytes
*/
void traceParameterBytes(const ParameterBytes& parameterBytes)
{
	std::cout.fill('0');
	std::cout << std::hex << std::uppercase;
	for (auto& item : parameterBytes)
	{
		std::cout << std::setw(2) << static_cast<int>(item) << " ";
	}
	std::cout << std::nouppercase << std::dec << std::endl;
}

}  // end anonymous namespace

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "02_BaosParameterBytes sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);

		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}

		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;

		// create a TPC/IP connection with the remote BAOS device
		BaosIp4Device baosDevice(ipAddress);

		// Reads some parameter bytes

		// index = 2; number of = 4
		ParameterBytes parameterBytes = baosDevice.getParameterBytes(2, 4);
		std::cout << "Parameter data (2 .. 5): ";
		traceParameterBytes(parameterBytes);
		std::cout << std::endl;

		// Reads all parameter bytes

		parameterBytes = baosDevice.getAllParameterBytes();
		std::cout << "Parameter data (" << parameterBytes.size() << " Bytes):";
		traceParameterBytes(parameterBytes);
		std::cout << std::endl;
	}
	catch (const wzcpp::error::WzBaseException& e)
	{
		std::cout << "Wz Exception in main: " << e.msg() << std::endl;
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
