//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include "Helper.h"

#include <baos/Sdk.h>
#include <baos/ip/BaosIpv4Device.h>
#include <baos/ip/BaosTcpConnection.h>
#include <baos/protocol/Defines.h>

#include <iostream>
#include <exception>
#include <memory>

/*! 
	@example 08_Counter.cpp
    
	This sample shows how to read a parameter and write datapoint values via the device API.
	
    Usage:
    @code
        08_Counter <ip_address>
        // e.g.:
        08_Counter 10.0.0.102
    @endcode
	
	Pre-condition:
	It assumes that the following datapoints and parameters are configured:
	
	@verbatim

	Datapoints:
	|---------|-----------------|-----------------------------|-------------|---------------|
	|   Id    |  DPT            |  Name                       |  SizeInBit  |  Flags        |
	|         |                 |                             |             | C R W T U I   |
	|---------|-----------------|-----------------------------|-------------|---------------|
	|    9    | DPT 01 - 1 bit  | Do Add                      | 1           | C - W - - -   |
	|---------|-----------------|-----------------------------|-------------|---------------|
	|   10    | DPT 07          | Counter                     | 16          | C R W T U -   |
	|---------|-----------------|-----------------------------|-------------|---------------|
	
	Parameter:
	|---------|-----------------|-----------------------------|-------------|
	| Number  |  Size           |  Name                       |  Value      |
	|         |                 |                             |             |
	|---------|-----------------|-----------------------------|-------------|
	|   2     | 1 Byte 			| Step		                  | 10          |
	|---------|-----------------|-----------------------------|-------------|
	

	@endverbatim
*/

using utility::CommandLineOptions;
using wz::baos::setBaosLogLevel;
using namespace wz::baos;
using namespace wz::baos::ip;
using namespace wz::baos::protocol;

/******************************************
** Anonymous namespace
*******************************************/

namespace
{

enum Datapoints : std::uint16_t
{
	DoAdd = 9, /*!< Datapoint id for datapoint "do add" */
	Counter  = 11, /*!< Datapoint id for datapoint counter */
};

enum Parameters : std::uint16_t
{
	Step = 2 /*!< Parameter id for step */
};

// NOTE: Overflow not handled here 
void doAdd(BaosIp4Device& baosDevice)
{
	// Get parameter for step
	ParameterBytes parameterBytes = baosDevice.getParameterBytes(Parameters::Step, 1);
	std::uint8_t step = parameterBytes.at(0);
	
	// Get current value
	std::uint16_t counter = 0;
	auto maybeValue = baosDevice.getDpValueRaw(Datapoints::Counter);
	if (maybeValue.has_value())
	{
		counter = (maybeValue->at(0) << 8) | maybeValue->at(1);
	}

	// Calculate new value
	counter += step;

	// Write new value
	std::uint8_t highByte = (counter & 0xFF00) >> 8;
	std::uint8_t lowByte = counter & 0xFF;
	baosDevice.setDpValue(1, {highByte, lowByte}, DatapointSetCommand::SetNewValueAndSendOnBus);
}

}

/******************************************
** Main
*******************************************/

int main(int argc, char* argv[])
{
	try
	{
		std::cout << "******************************************" << std::endl;
		std::cout << "08_Counter sample" << std::endl;
		std::cout << "******************************************" << std::endl;
		std::cout << std::endl;

		// Reduce log messages
		setBaosLogLevel(wzcpp::LogLevel::warn);
		
		// The CommandLineOptions::parse function
		// validates the arguments and handles the help command
		CommandLineOptions options;
		options.parse(argc, argv);
		if (options.wasHelpdisplayed())
		{
			return 0;
		}
		
		// Get the ip address from the parsed command line options
		const std::string ipAddress = options.getIpAddress();
		std::cout << " Connection requested for " << ipAddress << std::endl;
		
		// create a TPC/IP connection to the remote BAOS device
		BaosIp4Device baosDevice(ipAddress);
		
		// read static server items
		// NOTE: This does an auto connect
		const DeviceStaticInfo staticInfo = baosDevice.getStaticInfo();

		// register callback function for datapoint value indications
		baosDevice.registerDPValueCallback(
		    [&baosDevice](const DatapointValues& values)
		{
			    for (const auto& item : values)
			    {
				    std::uint16_t id            = item.first;
				    const DatapointValue& value = item.second;

				    std::cout << "Received datapoint value indication for id " << id << std::endl;

				    switch (id)
				    {
					    case Datapoints::DoAdd:
						    std::cout << "Add step value to counter " << std::endl;
						    doAdd(baosDevice);
							break;
					    case Datapoints::Counter:
						    std::cout << "Counter was changed from other (bus or other baos client)" << std::endl;
						    break;
					    default:
						    std::cout << "Nothing to do" << std::endl;
				    }
			    }
		});
		
		// wait for Enter, all indications will be shown
		// in the datapoint event handler
		std::cout << "Press [Enter] to exit the application ..." << std::endl;
		std::getchar();	
	}
	catch (const std::exception& e)
	{
		std::cerr << "Failed: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}

